import re
import string
from intermine314.pathfeatures import PathFeature, PATH_PATTERN
from intermine314.util import ReadableException


class Constraint(PathFeature):
    """
    A class representing constraints on a query
    ===========================================

    All constraints inherit from this class, which
    simply defines the type of element for the
    purposes of serialisation.
    """

    child_type = "constraint"


class LogicNode(object):
    """
    A class representing nodes in a logic graph
    ===========================================

    Objects which can be represented as nodes
    in the AST of a constraint logic graph should
    inherit from this class, which defines
    methods for overloading built-in operations.
    """

    def __add__(self, other):
        """
        Overloads +
        ===========

        Logic may be defined by using addition to sum
        logic nodes::

            > query.set_logic(con_a + con_b + con_c)
            > str(query.logic)
            ... A and B and C

        """
        if not isinstance(other, LogicNode):
            return NotImplemented
        else:
            return LogicGroup(self, "AND", other)

    def __and__(self, other):
        """
        Overloads &
        ===========

        Logic may be defined by using the & operator::

            > query.set_logic(con_a & con_b)
            > sr(query.logic)
            ... A and B

        """
        if not isinstance(other, LogicNode):
            return NotImplemented
        else:
            return LogicGroup(self, "AND", other)

    def __or__(self, other):
        """
        Overloads |
        ===========

        Logic may be defined by using the | operator::

            > query.set_logic(con_a | con_b)
            > str(query.logic)
            ... A or B

        """
        if not isinstance(other, LogicNode):
            return NotImplemented
        else:
            return LogicGroup(self, "OR", other)


class LogicGroup(LogicNode):
    """
    A logic node that represents two sub-nodes joined in some way
    =============================================================

    A logic group is a logic node with two child nodes, which are
    either connected by AND or by OR logic.
    """

    LEGAL_OPS = frozenset(["AND", "OR"])

    def __init__(self, left, op, right, parent=None):
        """
        Constructor
        ===========

        Makes a new node composes of two nodes (left and right),
        and some operator.

        Groups may have a reference to their parent.
        """
        if op not in self.LEGAL_OPS:
            raise TypeError(op + " is not a legal logical operation")
        self.parent = parent
        self.left = left
        self.right = right
        self.op = op
        for node in [self.left, self.right]:
            if isinstance(node, LogicGroup):
                node.parent = self

    def __repr__(self):
        """
        Provide a sensible representation of a node
        """
        return "<" + self.__class__.__name__ + ": " + str(self) + ">"

    def __str__(self):
        """
        Provide a human readable version of the group. The
        string version should be able to be parsed back into the
        original logic group.
        """
        core = " ".join(map(str, [self.left, self.op.lower(), self.right]))
        if self.parent and self.op != self.parent.op:
            return "(" + core + ")"
        else:
            return core

    def get_codes(self):
        """
        Get a list of all constraint codes used in this group.
        """
        codes = []
        for node in [self.left, self.right]:
            if isinstance(node, LogicGroup):
                codes.extend(node.get_codes())
            else:
                codes.append(node.code)
        return codes


class LogicParseError(ReadableException):
    """
    An error representing problems in parsing constraint logic.
    """

    pass


class EmptyLogicError(ValueError):
    """
    An error representing the fact that an the logic
    string to be parsed was empty
    """

    pass


class LogicParser(object):
    """
    Parses logic strings into logic groups
    ======================================

    Instances of this class are used to parse logic strings into
    abstract syntax trees, and then logic groups. This aims to provide
    robust parsing of logic strings, with the ability to identify syntax
    errors in such strings.
    """

    def __init__(self, query):
        """
        Constructor
        ===========

        Parsers need access to the query they are parsing for, in
        order to reference the constraints on the query.

        @param query: The parent query object
        @type query: intermine314.query.Query
        """
        self._query = query

    def get_constraint(self, code):
        """
        Get the constraint with the given code
        ======================================

        This method fetches the constraint from the
        parent query with the matching code.

        @see: intermine314.query.Query.get_constraint
        @rtype: intermine314.constraints.CodedConstraint
        """
        return self._query.get_constraint(code)

    def get_priority(self, op):
        """
        Get the priority for a given operator
        =====================================

        Operators have a specific precedence, from highest
        to lowest:
          - ()
          - AND
          - OR

        This method returns an integer which can be
        used to compare operator priorities.

        @rtype: int
        """
        return {"AND": 2, "OR": 1, "(": 3, ")": 3}.get(op)

    ops = {"AND": "AND", "&": "AND", "&&": "AND", "OR": "OR", "|": "OR", "||": "OR", "(": "(", ")": ")"}

    def parse(self, logic_str):
        """
        Parse a logic string into an abstract syntax tree
        =================================================

        Takes a string such as "A and B or C and D", and parses it
        into a structure which represents this logic as a binary
        abstract syntax tree. The above string would parse to
        "(A and B) or (C and D)", as AND binds more tightly than OR.

        Note that only singly rooted trees are parsed.

        @param logic_str: The logic defininition as a string
        @type logic_str: string

        @rtype: LogicGroup

        @raise LogicParseError: if there is a syntax error in the logic
        """

        def flatten(l):
            """Flatten out a list which contains both values and sublists"""
            ret = []
            for item in l:
                if isinstance(item, list):
                    ret.extend(item)
                else:
                    ret.append(item)
            return ret

        def canonical(x, d):
            if x in d:
                return d[x]
            else:
                return re.split("\b", x)

        def dedouble(x):
            if re.search("[()]", x):
                return list(x)
            else:
                return x

        logic_str = logic_str.upper()
        tokens = [t for t in re.split("\\s+", logic_str) if t]
        if not tokens:
            raise EmptyLogicError()
        tokens = flatten([canonical(x, self.ops) for x in tokens])
        tokens = flatten([dedouble(x) for x in tokens])
        self.check_syntax(tokens)
        postfix_tokens = self.infix_to_postfix(tokens)
        abstract_syntax_tree = self.postfix_to_tree(postfix_tokens)
        return abstract_syntax_tree

    def check_syntax(self, infix_tokens):
        """
        Check the syntax for errors before parsing
        ==========================================

        Syntax is checked before parsing to provide better errors,
        which should hopefully lead to more informative error messages.

        This checks for:
         - correct operator positions (cannot put two codes next to each
           other without intervening operators)
         - correct grouping (all brackets are matched,
                             and contain valid expressions)

        @param infix_tokens: The input parsed into a list of tokens.
        @type infix_tokens: iterable

        @raise LogicParseError: if there is a problem.
        """
        need_an_op = False
        need_binary_op_or_closing_bracket = False
        processed = []
        open_brackets = 0
        for token in infix_tokens:
            if token not in self.ops:
                if need_an_op:
                    raise LogicParseError(
                        "Expected an operator after: '" + " ".join(processed) + "'" + " - but got: '" + token + "'"
                    )
                if need_binary_op_or_closing_bracket:
                    raise LogicParseError(
                        "Logic grouping error after: '" + " ".join(processed) + "'" + " - expected an operator "
                        "or a closing bracket"
                    )

                need_an_op = True
            else:
                need_an_op = False
                if token == "(":
                    if processed and processed[-1] not in self.ops:
                        raise LogicParseError(
                            "Logic grouping error after: '" + " ".join(processed) + "'" + " - got an unexpeced "
                            "opening bracket"
                        )
                    if need_binary_op_or_closing_bracket:
                        raise LogicParseError(
                            "Logic grouping error after: '" + " ".join(processed) + "'" + " - expected an operator or "
                            "a closing bracket"
                        )

                    open_brackets += 1
                elif token == ")":
                    need_binary_op_or_closing_bracket = True
                    open_brackets -= 1
                else:
                    need_binary_op_or_closing_bracket = False
            processed.append(token)
        if open_brackets != 0:
            if open_brackets < 0:
                message = "Unmatched closing bracket in: "
            else:
                message = "Unmatched opening bracket in: "
            raise LogicParseError(message + '"' + " ".join(infix_tokens) + '"')

    def infix_to_postfix(self, infix_tokens):
        """
        Convert a list of infix tokens to postfix notation
        ==================================================

        Take in a set of infix tokens and return the set parsed
        to a postfix sequence.

        @param infix_tokens: The list of tokens
        @type infix_tokens: iterable

        @rtype: list
        """
        stack = []
        postfix_tokens = []
        for token in infix_tokens:
            if token not in self.ops:
                postfix_tokens.append(token)
            else:
                op = token
                if op == "(":
                    stack.append(token)
                elif op == ")":
                    while stack:
                        last_op = stack.pop()
                        if last_op == "(":
                            if stack:
                                previous_op = stack.pop()
                                if previous_op != "(":
                                    postfix_tokens.append(previous_op)
                                break
                        else:
                            postfix_tokens.append(last_op)
                else:
                    while stack and (self.get_priority(stack[-1]) <= self.get_priority(op)):
                        prev_op = stack.pop()
                        if prev_op != "(":
                            postfix_tokens.append(prev_op)
                    stack.append(op)
        while stack:
            postfix_tokens.append(stack.pop())
        return postfix_tokens

    def postfix_to_tree(self, postfix_tokens):
        """
        Convert a set of structured tokens to a single LogicGroup
        =========================================================

        Convert a set of tokens in postfix notation to a single
        LogicGroup object.

        @param postfix_tokens: A list of tokens in postfix notation.
        @type postfix_tokens: list

        @rtype: LogicGroup

        @raise AssertionError: is the tree doesn't have a unique root.
        """
        stack = []
        try:
            for token in postfix_tokens:
                if token not in self.ops:
                    stack.append(self.get_constraint(token))
                else:
                    op = token
                    right = stack.pop()
                    left = stack.pop()
                    stack.append(LogicGroup(left, op, right))
            assert len(stack) == 1, "Tree doesn't have a unique root"
            return stack.pop()
        except IndexError:
            raise EmptyLogicError()


class CodedConstraint(Constraint, LogicNode):
    """
    A parent class for all constraints that have codes
    ==================================================

    Constraints that have codes are the principal logical
    filters on queries, and need to be refered to individually
    (hence the codes). They will all have a logical operation they
    embody, and so have a reference to an operator.

    This class is not meant to be instantiated directly, but instead
    inherited from to supply default behaviour.
    """

    OPS = set([])

    def __init__(self, path, op, code="A"):
        """
        Constructor
        ===========

        @param path: The path to constrain
        @type path: string

        @param op: The operation to apply - must be in the OPS set
        @type op: string
        """
        if op not in self.OPS:
            raise TypeError(op + " not in " + str(self.OPS))
        self.op = op
        self.code = code
        super(CodedConstraint, self).__init__(path)

    def get_codes(self):
        return [self.code]

    def __str__(self):
        """
        Stringify to the code they are refered to by.
        """
        return self.code

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(CodedConstraint, self).to_string()
        return " ".join([s, self.op])

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(CodedConstraint, self).to_dict()
        d.update(op=self.op, code=self.code)
        return d


class UnaryConstraint(CodedConstraint):
    """
    Constraints which have just a path and an operator
    ==================================================

    These constraints are simple assertions about the
    object/value refered to by the path. The set of valid
    operators is:
     - IS NULL
     - IS NOT NULL

    """

    OPS = set(["IS NULL", "IS NOT NULL"])


class BinaryConstraint(CodedConstraint):
    """
    Constraints which have an operator and a value
    ==============================================

    These constraints assert a relationship between the
    value represented by the path (it must be a representation
    of a value, ie an Attribute) and another value - ie. the
    operator takes two parameters.

    In all case the 'left' side of the relationship is the path,
    and the 'right' side is the supplied value.

    Valid operators are:
     - =        (equal to)
     - !=       (not equal to)
     - <        (less than)
     - >        (greater than)
     - <=       (less than or equal to)
     - >=       (greater than or equal to)
     - LIKE     (same as equal to, but with implied wildcards)
     - CONTAINS (same as equal to, but with implied wildcards)
     - NOT LIKE (same as not equal to, but with implied wildcards)

    """

    OPS = set(["=", "!=", "<", ">", "<=", ">=", "LIKE", "NOT LIKE", "CONTAINS"])

    def __init__(self, path, op, value, code="A"):
        """
        Constructor
        ===========

        @param path: The path to constrain
        @type path: string

        @param op: The relationship between the value represented by the path
                   and the value provided (must be a valid operator)
        @type op: string

        @param value: The value to compare the stored value to
        @type value: string or number

        @param code: The code for this constraint (default = "A")
        @type code: string
        """
        self.value = value
        super(BinaryConstraint, self).__init__(path, op, code)

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(BinaryConstraint, self).to_string()
        return " ".join([s, str(self.value)])

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(BinaryConstraint, self).to_dict()
        d.update(value=str(self.value))
        return d


class ListConstraint(CodedConstraint):
    """
    Constraints which refer to an objects membership of lists
    =========================================================

    These constraints assert a membership relationship between the
    object represented by the path (it must always be an object, ie.
    a Reference or a Class) and a List. Lists are collections of
    objects in the database which are stored in InterMine
    datawarehouses. These lists must be set up before the query is run, either
    manually in the webapp or by using the webservice API list
    upload feature.

    Valid operators are:
     - IN
     - NOT IN

    """

    OPS = set(["IN", "NOT IN"])

    def __init__(self, path, op, list_name, code="A"):
        if hasattr(list_name, "to_query"):
            q = list_name.to_query()
            list_name1 = q.service.create_list(q)
            self.list_name = list_name1.name
        elif hasattr(list_name, "name"):
            self.list_name = list_name.name
        else:
            self.list_name = list_name
        super(ListConstraint, self).__init__(path, op, code)

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(ListConstraint, self).to_string()
        return " ".join([s, str(self.list_name)])

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(ListConstraint, self).to_dict()
        d.update(value=str(self.list_name))
        return d


class LoopConstraint(CodedConstraint):
    """
    Constraints with refer to object identity
    =========================================

    These constraints assert that two paths refer to the same
    object.

    Valid operators:
     - IS
     - IS NOT

    The operators IS and IS NOT map to the ops "=" and "!=" when they
    are used in XML serialisation.

    """

    OPS = set(["IS", "IS NOT"])
    SERIALISED_OPS = {"IS": "=", "IS NOT": "!="}

    def __init__(self, path, op, loopPath, code="A"):
        """
        Constructor
        ===========

        @param path: The path to constrain
        @type path: string

        @param op: The relationship between the path and the path provided
                   (must be a valid operator)
        @type op: string

        @param loopPath: The path to check for identity against
        @type loopPath: string

        @param code: The code for this constraint (default = "A")
        @type code: string
        """
        self.loopPath = loopPath
        super(LoopConstraint, self).__init__(path, op, code)

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(LoopConstraint, self).to_string()
        return " ".join([s, self.loopPath])

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(LoopConstraint, self).to_dict()
        d.update(loopPath=self.loopPath, op=self.SERIALISED_OPS[self.op])
        return d


class TernaryConstraint(BinaryConstraint):
    """
    Constraints for broad, general searching over all fields
    ========================================================

    These constraints request a wide-ranging search for matching
    fields over all aspects of an object, including up to coercion
    from related classes.

    Valid operators:
     - LOOKUP

    To aid disambiguation, Ternary constaints accept an extra_value as
    well as the main value.
    """

    OPS = set(["LOOKUP"])

    def __init__(self, path, op, value, extra_value=None, code="A"):
        """
        Constructor
        ===========

        @param path: The path to constrain. Here is must be a class,
                     or a reference to a class.
        @type path: string

        @param op: The relationship between the path and the path provided
                   (must be a valid operator)
        @type op: string

        @param value: The value to check other fields against.
        @type value: string

        @param extra_value: A further value for disambiguation. The meaning
                            of this value varies by class and configuration.
                            For example, if the class of the object is Gene,
                            then extra_value will refer to the Organism.
        @type extra_value: string

        @param code: The code for this constraint (default = "A")
        @type code: string
        """
        self.extra_value = extra_value
        super(TernaryConstraint, self).__init__(path, op, value, code)

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(TernaryConstraint, self).to_string()
        if self.extra_value is None:
            return s
        else:
            return " ".join([s, "IN", self.extra_value])

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(TernaryConstraint, self).to_dict()
        if self.extra_value is not None:
            d.update(extraValue=self.extra_value)
        return d


class MultiConstraint(CodedConstraint):
    """
    Constraints for checking membership of a set of values
    ======================================================

    These constraints require the value they constrain to be
    either a member of a set of values, or not a member.

    Valid operators:
     - ONE OF
     - NONE OF

    These constraints are similar in use to List constraints, with
    the following differences:
      - The list in this case is a defined set of values that is passed
        along with the query itself, rather than anything stored
        independently on a server.
      - The object of the constaint is the value of an attribute, rather
        than an object's identity.
    """

    OPS = set(["ONE OF", "NONE OF"])

    def __init__(self, path, op, values, code="A"):
        """
        Constructor
        ===========

        @param path: The path to constrain. Here it must be an attribute of
                     some object.
        @type path: string

        @param op: The relationship between the path and the path provided
                   (must be a valid operator)
        @type op: string

        @param values: The set of values which the object of the constraint
                       either must or must not belong to.
        @type values: set or list

        @param code: The code for this constraint (default = "A")
        @type code: string
        """
        if not isinstance(values, (set, list)):
            raise TypeError("values must be a set or a list, not " + str(type(values)))
        self.values = values
        super(MultiConstraint, self).__init__(path, op, code)

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(MultiConstraint, self).to_string()
        return " ".join([s, str(self.values)])

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(MultiConstraint, self).to_dict()
        d.update(value=self.values)
        return d


class RangeConstraint(MultiConstraint):
    """
    Constraints for testing where a value lies relative to a set of ranges
    ======================================================================

    These constraints require that the value of the path they constrain
    should lie in relationship to the set of values passed according to
    the specific operator.

    Valid operators:
     - OVERLAPS : The value overlaps at least one of the given ranges
     - WITHIN : The value is wholly outside the given set of ranges
     - CONTAINS : The value contains all the given ranges
     - DOES NOT CONTAIN : The value does not contain all the given ranges
     - OUTSIDE : Some part is outside the given set of ranges
     - DOES NOT OVERLAP : The value does not overlap with any of the ranges

    For example:

        4 WITHIN [1..5, 20..25] => True

    The format of the ranges depends on the value being constrained and what
    range parsers have been configured on the target server. A common range
    parser for biological mines is the one for Locations:

        Gene.chromosomeLocation OVERLAPS [2X:54321..67890, 3R:12345..456789]

    """

    OPS = set(["OVERLAPS", "DOES NOT OVERLAP", "WITHIN", "OUTSIDE", "CONTAINS", "DOES NOT CONTAIN"])


class IsaConstraint(MultiConstraint):
    """
    Constraints for testing the class of a value, as a disjoint union
    ======================================================================

    These constraints require that the value of the path they constrain
    should be an instance of one of the classes provided.

    Valid operators:
     - ISA : The value is an instance of one of the provided classes.

    For example:

        SequenceFeature ISA [Exon, Intron]

    """

    OPS = set(["ISA"])


class SubClassConstraint(Constraint):
    """
    Constraints on the class of a reference
    =======================================

    If an object has a reference X to another object of type A,
    and type B extends type A, then any object of type B may be
    the value of the reference X. If you only want to see X's
    which are B's, this may be achieved with subclass constraints,
    which allow the type of an object to be limited to one of the
    subclasses (at any depth) of the class type required
    by the attribute.

    These constraints do not use operators. Since they cannot be
    conditional (eg. "A is a B or A is a C" would not be possible
    in an InterMine query), they do not have codes
    and cannot be referenced in logic expressions.
    """

    def __init__(self, path, subclass):
        """
        Constructor
        ===========

        @param path: The path to constrain. This must refer to a class or a
                     reference to a class.
        @type path: str

        @param subclass: The class to subclass the path to. This must be a
                         simple class name (not a dotted name)
        @type subclass: str
        """
        if not PATH_PATTERN.match(subclass):
            raise TypeError
        self.subclass = subclass
        super(SubClassConstraint, self).__init__(path)

    def to_string(self):
        """
        Provide a human readable representation of the logic.
        This method is called by repr.
        """
        s = super(SubClassConstraint, self).to_string()
        return s + " ISA " + self.subclass

    def to_dict(self):
        """
        Return a dict object which can be used to construct a
        DOM element with the appropriate attributes.
        """
        d = super(SubClassConstraint, self).to_dict()
        d.update(type=self.subclass)
        return d


class TemplateConstraint(object):
    """
    A mixin to supply the behaviour and state of constraints on templates
    =====================================================================

    Constraints on templates can also be designated as "on", "off" or
    "locked", which refers to whether they are active or not. Inactive
    constraints are still configured, but behave as if absent for the purpose
    of results. In addition, template constraints can be editable or not.
    Only values for editable constraints can be provided when requesting
    results, and only constraints that can participate in logic expressions
    can be editable.
    """

    REQUIRED = "locked"
    OPTIONAL_ON = "on"
    OPTIONAL_OFF = "off"

    def __init__(self, editable=True, optional="locked"):
        """
        Constructor
        ===========

        @param editable: Whether or not this constraint should accept new
                         values.
        @type editable: bool

        @param optional: Whether a value for this constraint must be provided
                         when running.
        @type optional: "locked", "on" or "off"
        """
        self.editable = editable
        if optional == TemplateConstraint.REQUIRED:
            self.optional = False
            self.switched_on = True
        else:
            self.optional = True
            if optional == TemplateConstraint.OPTIONAL_ON:
                self.switched_on = True
            elif optional == TemplateConstraint.OPTIONAL_OFF:
                self.switched_on = False
            else:
                raise TypeError("Bad value for optional")

    @property
    def required(self):
        """
        True if a value must be provided for this constraint.

        @rtype: bool
        """
        return not self.optional

    @property
    def switched_off(self):
        """
        True if this constraint is currently inactive.

        @rtype: bool
        """
        return not self.switched_on

    def get_switchable_status(self):
        """
        Returns either "locked", "on" or "off".
        """
        if not self.optional:
            return "locked"
        else:
            if self.switched_on:
                return "on"
            else:
                return "off"

    def switch_on(self):
        """
        Make sure this constraint is active
        ===================================

        @raise ValueError: if the constraint is not editable and optional
        """
        if self.editable and self.optional:
            self.switched_on = True
        else:
            raise ValueError("This constraint is not switchable")

    def switch_off(self):
        """
        Make sure this constraint is inactive
        =====================================

        @raise ValueError: if the constraint is not editable and optional
        """
        if self.editable and self.optional:
            self.switched_on = False
        else:
            raise ValueError("This constraint is not switchable")

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        if self.editable:
            editable = "editable"
        else:
            editable = "non-editable"
        return "(" + editable + ", " + self.get_switchable_status() + ")"

    def separate_arg_sets(self, args):
        """
        A static function to use when building template constraints.
        ============================================================

        dict -> (dict, dict)

        Splits a dictionary of arguments into two separate dictionaries, one
        with arguments for the main constraint, and one with arguments for the
        template portion of the behaviour
        """
        c_args = {}
        t_args = {}
        for k, v in list(args.items()):
            if k == "editable":
                t_args[k] = v == "true"
            elif k == "optional":
                t_args[k] = v
            else:
                c_args[k] = v
        return (c_args, t_args)


class TemplateUnaryConstraint(UnaryConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        UnaryConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return UnaryConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateBinaryConstraint(BinaryConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        BinaryConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return BinaryConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateListConstraint(ListConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        ListConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return ListConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateLoopConstraint(LoopConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        LoopConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return LoopConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateTernaryConstraint(TernaryConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        TernaryConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return TernaryConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateMultiConstraint(MultiConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        MultiConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return MultiConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateRangeConstraint(RangeConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        RangeConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return RangeConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateIsaConstraint(IsaConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        IsaConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return IsaConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class TemplateSubClassConstraint(SubClassConstraint, TemplateConstraint):
    def __init__(self, *a, **d):
        (c_args, t_args) = self.separate_arg_sets(d)
        SubClassConstraint.__init__(self, *a, **c_args)
        TemplateConstraint.__init__(self, **t_args)

    def to_string(self):
        """
        Provide a template specific human readable representation of the
        constraint. This method is called by repr.
        """
        return SubClassConstraint.to_string(self) + " " + TemplateConstraint.to_string(self)


class ConstraintFactory(object):
    """
    A factory for creating constraints from a set of arguments.
    ===========================================================

    A constraint factory is responsible for finding an appropriate
    constraint class for the given arguments and instantiating the
    constraint.
    """

    CONSTRAINT_CLASSES = set(
        [
            UnaryConstraint,
            BinaryConstraint,
            TernaryConstraint,
            MultiConstraint,
            SubClassConstraint,
            LoopConstraint,
            ListConstraint,
            RangeConstraint,
            IsaConstraint,
        ]
    )

    def __init__(self):
        """
        Constructor
        ===========

        Creates a new ConstraintFactory
        """
        self._codes = iter(string.ascii_uppercase)
        self.reference_ops = TernaryConstraint.OPS | RangeConstraint.OPS | ListConstraint.OPS | IsaConstraint.OPS

    def get_next_code(self):
        """
        Return the available constraint code.

        @return: A single uppercase character
        @rtype: str
        """
        return next(self._codes)

    def make_constraint(self, *args, **kwargs):
        """
        Create a constraint from a set of arguments.
        ============================================

        Finds a suitable constraint class, and instantiates it.

        @rtype: Constraint
        """
        for CC in self.CONSTRAINT_CLASSES:
            try:
                c = CC(*args, **kwargs)
                if hasattr(c, "code") and c.code == "A":
                    c.code = self.get_next_code()
                return c
            except TypeError as e:
                pass
        raise TypeError("No matching constraint class found for " + str(args) + ", " + str(kwargs))


class TemplateConstraintFactory(ConstraintFactory):
    """
    A factory for creating constraints with template specific characteristics.
    ==========================================================================

    A constraint factory is responsible for finding an appropriate
    constraint class for the given arguments and instantiating the
    constraint. TemplateConstraintFactories make constraints with the
    extra set of TemplateConstraint qualities.
    """

    CONSTRAINT_CLASSES = set(
        [
            TemplateUnaryConstraint,
            TemplateBinaryConstraint,
            TemplateTernaryConstraint,
            TemplateMultiConstraint,
            TemplateSubClassConstraint,
            TemplateLoopConstraint,
            TemplateListConstraint,
            TemplateRangeConstraint,
            TemplateIsaConstraint,
        ]
    )



import codecs
import json
import logging
import weakref
from contextlib import closing
from functools import partial
from urllib.parse import urlencode

from intermine314.errors import WebserviceError
from intermine314.lists.list import List

logging.basicConfig()


class ListManager:
    """
    A Class for Managing List Content and Operations
    ================================================

    This class provides methods to manage list contents and operations.

    This class may be called itself, but all the useful methods it has
    are also available on the Service object, which delegates to this class,
    while other methods are more coneniently accessed through the list objects
    themselves.

    NB: The methods for creating lists can conflict in threaded applications,
    if two threads are each allocated the same unused list name. You are
    strongly advised to use locks to synchronise any list creation requests
    (create_list, or intersect, union, subtract, diff) unless you are choosing
    your own names each time and are confident that these will not conflict.
    """

    LOG = logging.getLogger("listmanager")
    DEFAULT_LIST_NAME = "my_list"
    DEFAULT_DESCRIPTION = "List created with Python client library"

    INTERSECTION_PATH = "/lists/intersect/json"
    UNION_PATH = "/lists/union/json"
    DIFFERENCE_PATH = "/lists/diff/json"
    SUBTRACTION_PATH = "/lists/subtract/json"

    def __init__(self, service):
        self.service = weakref.proxy(service)
        self.lists = None
        self._temp_lists = set()

    def refresh_lists(self):
        """
        Update the list information with the latest details from the server
        """

        self.lists = {}
        url = self.service.root + self.service.LIST_PATH
        data = self.service.opener.read(url)
        list_info = json.loads(data)
        self.LOG.debug("LIST INFO: {0}".format(list_info))
        if not list_info.get("wasSuccessful"):
            raise ListServiceError(list_info.get("error"))
        for l in list_info["lists"]:
            l = ListManager.safe_dict(l)
            self.lists[l["name"]] = List(service=self.service, manager=self, **l)

    @staticmethod
    def safe_dict(d):
        """Recursively clone json structure with UTF-8 dictionary keys"""

        if isinstance(d, dict):
            return dict(d)
        return d

    def get_list(self, name):
        """Return a list from the service by name, if it exists"""

        if self.lists is None:
            self.refresh_lists()
        return self.lists.get(name)

    def l(self, name):
        """Alias for get_list"""

        return self.get_list(name)

    def get_all_lists(self):
        """Get all the lists on a webservice"""

        if self.lists is None:
            self.refresh_lists()
        return self.lists.values()

    def get_all_list_names(self):
        """Get all the names of the lists in a particular webservice"""

        if self.lists is None:
            self.refresh_lists()
        return self.lists.keys()

    def get_list_count(self):
        """
        Return the number of lists accessible at the given webservice.
        This number will vary depending on who you are authenticated as.
        """

        return len(self.get_all_list_names())

    def get_unused_list_name(self):
        """
        Get an unused list name
        =======================

        This method returns a new name that does not conflict
        with any currently existing list name.

        The list name is only guaranteed to be unused at the time
        of allocation.
        """

        self.refresh_lists()
        list_names = self.get_all_list_names()
        self.LOG.debug("CURRENT LIST NAMES: {0}".format(list_names))
        counter = 1
        get_name = partial("{0}_{1}".format, self.DEFAULT_LIST_NAME)
        name = get_name(counter)
        while name in list_names:
            counter += 1
            name = get_name(counter)
        self._temp_lists.add(name)
        return name

    def _get_listable_query(self, queryable):
        q = queryable.to_query()
        if not q.views:
            q.add_view(q.root.name + ".id")
        else:
            # Check to see if the class of the selected items is
            # unambiguous

            up_to_attrs = set(v[0 : v.rindex(".")] for v in q.views)
            if len(up_to_attrs) == 1:
                q.select(up_to_attrs.pop() + ".id")
        return q

    def _create_list_from_queryable(
        self,
        queryable,
        name,
        description,
        tags,
    ):
        q = self._get_listable_query(queryable)
        uri = q.get_list_upload_uri()
        params = q.to_query_params()
        params["listName"] = name
        params["description"] = description
        params["tags"] = ";".join(tags)
        form = urlencode(params)
        resp = self.service.opener.open(uri, form)
        data = resp.read()
        resp.close()
        return self.parse_list_upload_response(data)

    def create_list(
        self,
        content,
        list_type="",
        name=None,
        description=None,
        tags=None,
        add=None,
        organism=None,
    ):
        """
        Create a new list in the webservice
        ===================================

        If no name is given, the list will be considered to be a temporary
        list, and will be automatically deleted when the program ends.
        To prevent this happening, give the list a name, either on creation,
        or by renaming it.

        This method is not thread safe for anonymous lists - it will need
        synchronisation with locks if you intend to create lists with multiple
        threads in parallel.

        @param content: The source of the identifiers for this list.
        This can be:
                  * A string with white-space separated terms.
                  * The name of a file that contains the terms.
                  * A file-handle like thing (something with a 'read' method)
                  * An iterable of identifiers
                  * A query with a single column.
                  * Another list.
        @param list_type: The type of objects to include in the list.
                          This parameter is not required if the content
                          parameter implicitly includes the type
                          (as queries and lists do).
        @param name: The name for the new list.
                     If none is provided one will be generated, and the
                     list will be deleted when the list manager exits context.
        @param description: A description for the list
                            (free text, default = None)
        @param tags: A set of strings to use as tags (default = [])
        @param add: The issues groups that can be treated as matches.
                    This should be a collection of strings naming issue groups
                    that would otherwise be ignored, but in this case will be
                    added to the list. The available groups are:
                      * DUPLICATE - More than one match was found.
                      * WILDCARD - A wildcard match was made.
                      * TYPE_CONVERTED - A match was found, but in another type
                                         (eg. found a protein
                                         and we could convert it to a gene).
                      * OTHER - other issue types
                      * :all - All issues should be considered acceptable.
                    This only makes sense with text uploads
                    - it is not required (or used) when
                    the content is a list or a query.
        @param organism: organism name

        @rtype: intermine314.lists.List
        """

        if description is None:
            description = self.DEFAULT_DESCRIPTION
        if tags is None:
            tags = []
        if add is None:
            add = []

        if name is None:
            name = self.get_unused_list_name()
        if len(content) <= 0:
            print("Lists must have one or more elements - the current list has 0")
            print("Please create a valid list with at least one element and create the list again.")
            return

        item_content = content

        if organism:
            # If an organism name is given, create a query

            from intermine314.webservice import Service

            service = Service(self.service.root)
            query = service.new_query(list_type)

            # add organism constraint to the query

            if isinstance(organism, list):
                query.add_constraint("{0}.organism.name".format(list_type), "ONE OF", organism)
            else:
                query.add_constraint("organism", "LOOKUP", organism)
            if isinstance(item_content, list):
                # If symbols are given

                query.add_constraint("symbol", "ONE OF", item_content)

            # If one wants to create a list while
            # specifying an organism, then a content should not be
            # passed.

            item_content = query

        try:
            ids = item_content.read()  # File like thing
        except AttributeError:
            try:
                with closing(codecs.open(item_content, "r", "UTF-8")) as c:  # File name
                    ids = c.read()
            except (TypeError, IOError):
                try:
                    ids = item_content.strip()  # Stringy thing
                except AttributeError:
                    try:  # Queryable
                        return self._create_list_from_queryable(item_content, name, description, tags)
                    except AttributeError:
                        try:  # Array of idents
                            idents = iter(item_content)
                            ids = "\n".join(map('"{0}"'.format, idents))
                        except AttributeError:
                            raise TypeError("Cannot create list from " + repr(item_content))

        uri = self.service.root + self.service.LIST_CREATION_PATH
        query_form = {
            "name": name,
            "type": list_type,
            "description": description,
            "tags": ";".join(tags),
        }
        if len(add):
            query_form["add"] = [x.lower() for x in add if x]

        uri += "?" + urlencode(query_form, doseq=True)
        data = self.service.opener.post_plain_text(uri, ids)
        return self.parse_list_upload_response(data)

    def parse_list_upload_response(self, response):
        """
        Intepret the response from the webserver to a list request,
        and return the List it describes
        """

        try:
            response_data = json.loads(response.decode("utf8"))
        except ValueError:
            raise ListServiceError("Error parsing response: " + response)

        if not response_data.get("wasSuccessful"):
            raise ListServiceError(response_data.get("error"))

        self.LOG.debug("response data: {0}".format(response_data))
        self.refresh_lists()
        new_list = self.get_list(response_data["listName"])
        failed_matches = response_data.get("unmatchedIdentifiers")
        new_list._add_failed_matches(failed_matches)
        return new_list

    def delete_lists(self, lists):
        """Delete the given lists from the webserver"""

        self.refresh_lists()
        all_names = self.get_all_list_names()
        for l in lists:
            if isinstance(l, List):
                name = l.name
            else:
                name = str(l)
            if name not in all_names:
                self.LOG.debug("{0} does not exist - skipping".format(name))
                continue
            self.LOG.debug("deleting {0}".format(name))
            uri = self.service.root + self.service.LIST_PATH
            query_form = {"name": name}
            uri += "?" + urlencode(query_form)
            response = self.service.opener.delete(uri)
            response_data = json.loads(response.decode("utf8"))
            if not response_data.get("wasSuccessful"):
                raise ListServiceError(response_data.get("error"))
        self.refresh_lists()

    def remove_tags(self, to_remove_from, tags):
        """
        Add the tags to the given list
        ==============================

        Returns the current tags of this list.
        """

        uri = self.service.root + self.service.LIST_TAG_PATH
        form = {"name": to_remove_from.name, "tags": ";".join(tags)}
        uri += "?" + urlencode(form)
        body = self.service.opener.delete(uri)
        return self._body_to_json(body)["tags"]

    def add_tags(self, to_tag, tags):
        """
        Add the tags to the given list
        ==============================

        Returns the current tags of this list.
        """

        uri = self.service.root + self.service.LIST_TAG_PATH
        form = {"name": to_tag.name, "tags": ";".join(tags)}
        resp = self.service.opener.open(uri, urlencode(form))
        body = resp.read()
        resp.close()
        return self._body_to_json(body)["tags"]

    def get_tags(self, im_list):
        """
        Get the up-to-date set of tags for a given list
        ===============================================

        Returns the current tags of this list.
        """

        uri = self.service.root + self.service.LIST_TAG_PATH
        form = {"name": im_list.name}
        uri += "?" + urlencode(form)
        resp = self.service.opener.open(uri)
        body = resp.read()
        resp.close()
        return self._body_to_json(body)["tags"]

    def _body_to_json(self, body):
        try:
            data = json.loads(body.decode("utf8"))
        except ValueError:
            raise ListServiceError("Error parsing response: " + body)
        if not data.get("wasSuccessful"):
            raise ListServiceError(data.get("error"))
        return data

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type,
        exc_val,
        traceback,
    ):
        self.LOG.debug("Exiting context - deleting {0}".format(self._temp_lists))
        self.delete_temporary_lists()

    def delete_temporary_lists(self):
        """
        Delete all the lists considered temporary (those created without names)
        """

        if self._temp_lists:
            self.delete_lists(self._temp_lists)
            self._temp_lists = set()

    def intersect(
        self,
        lists,
        name=None,
        description=None,
        tags=None,
    ):
        """
        Calculate the intersection of a given set of lists, and return the
        list representing the result
        """

        return self._do_operation(
            self.INTERSECTION_PATH,
            "Intersection",
            lists,
            name,
            description,
            list(tags or []),
        )

    def union(
        self,
        lists,
        name=None,
        description=None,
        tags=None,
    ):
        """
        Calculate the union of a given set of lists,
        and return the list representing the result
        """

        return self._do_operation(
            self.UNION_PATH,
            "Union",
            lists,
            name,
            description,
            list(tags or []),
        )

    def xor(
        self,
        lists,
        name=None,
        description=None,
        tags=None,
    ):
        """
        Calculate the symmetric difference of a given set of lists,
        and return the list representing the result
        """

        return self._do_operation(
            self.DIFFERENCE_PATH,
            "Difference",
            lists,
            name,
            description,
            list(tags or []),
        )

    def subtract(
        self,
        lefts,
        rights,
        name=None,
        description=None,
        tags=None,
    ):
        """
        Calculate the subtraction of rights from lefts,
        and return the list representing the result
        """

        left_names = self.make_list_names(lefts)
        right_names = self.make_list_names(rights)
        tags = list(tags or [])
        if description is None:
            description = "Subtraction of " + " and ".join(right_names) + " from " + " and ".join(left_names)
        if name is None:
            name = self.get_unused_list_name()
        uri = self.service.root + self.SUBTRACTION_PATH
        uri += "?" + urlencode(
            {
                "name": name,
                "description": description,
                "references": ";".join(left_names),
                "subtract": ";".join(right_names),
                "tags": ";".join(tags),
            }
        )
        resp = self.service.opener.open(uri)
        data = resp.read()
        resp.close()
        return self.parse_list_upload_response(data)

    def _do_operation(
        self,
        path,
        operation,
        lists,
        name,
        description,
        tags,
    ):
        list_names = self.make_list_names(lists)
        if description is None:
            description = operation + " of " + " and ".join(list_names)
        if name is None:
            name = self.get_unused_list_name()
        uri = self.service.root + path
        uri += "?" + urlencode(
            {
                "name": name,
                "lists": ";".join(list_names),
                "description": description,
                "tags": ";".join(tags),
            }
        )
        resp = self.service.opener.open(uri)
        data = resp.read()
        resp.close()
        return self.parse_list_upload_response(data)

    def make_list_names(self, lists):
        """Turn a list of things into a list of list names"""

        list_names = []
        for l in lists:
            try:
                t = l.list_type
                list_names.append(l.name)
            except AttributeError:
                try:
                    m = l.model
                    list_names.append(self.create_list(l).name)
                except AttributeError:
                    list_names.append(str(l))

        return list_names


class ListServiceError(WebserviceError):
    """Errors thrown when something goes wrong with list requests"""

    pass



import weakref
import re
import logging
from xml.etree import ElementTree as ET
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union

from intermine314.util import openAnything, ReadableException

from functools import reduce

logging.basicConfig()

"""
Classes representing the data model
===================================

Representations of tables and columns, and behaviour
for validating connections between them.

"""

__author__ = "Alex Kalderimis"
__organization__ = "InterMine"
__license__ = "LGPL"
__contact__ = "toffe.kari@gmail.com"


class Field:
    """
    A class representing columns on database tables
    ===============================================

    The base class for attributes, references and collections. All
    columns in DB tables are represented by fields

    SYNOPSIS
    --------

        >>> service = Service("https://www.flymine.org/query/service")
        >>> model = service.model
        >>> cd = model.get_class("Gene")
        >>> print("Gene has", len(cd.fields), "fields")
        >>> for field in gene_cd.fields:
        ...        print(" - ", field)
        Gene has 45 fields
            -  CDSs is a group of CDS objects, which link back to this as gene
            -  GLEANRsymbol is a String
            -  UTRs is a group of UTR objects, which link back to this as gene
            -  alleles is a group of Allele objects, which link back
               to this as gene
            -  chromosome is a Chromosome
            -  chromosomeLocation is a Location
            -  clones is a group of CDNAClone objects, which link
               back to this as gene
            -  crossReferences is a group of CrossReference objects,
                which link back to this as subject
            -  cytoLocation is a String
            -  dataSets is a group of DataSet objects,
                which link back to this as bioEntities
            -  downstreamIntergenicRegion is a IntergenicRegion
            -  exons is a group of Exon objects,
                which link back to this as gene
            -  flankingRegions is a group of GeneFlankingRegion objects,
                which link back to this as gene
            -  goAnnotation is a group of GOAnnotation objects
            -  homologues is a group of Homologue objects,
                which link back to this as gene
            -  id is a Integer
            -  interactions is a group of Interaction objects,
                which link back to this as gene
            -  length is a Integer
            ...

    @see: L{Attribute}
    @see: L{Reference}
    @see: L{Collection}
    """

    def __init__(self, name: str, type_name: str, class_origin: "Class") -> None:
        """
        Constructor - DO NOT USE
        ========================

        THIS CLASS IS NOT MEANT TO BE INSTANTIATED DIRECTLY

        you are unlikely to need to do
        so anyway: it is recommended you access fields
        through the classes generated by the model

        @param name: The name of the reference
        @param type_name: The name of the model.Class this refers to
        @param class_origin: The model.Class this was declared in

        """
        self.name: str = name
        self.type_name: str = type_name
        self.type_class: Optional["Class"] = None
        self.declared_in: "Class" = class_origin

    def __repr__(self) -> str:
        return self.name + " is a " + self.type_name

    def __str__(self) -> str:
        return self.name

    @property
    def fieldtype(self) -> str:
        raise Exception("Fields should never be directly instantiated")


class Attribute(Field):
    """
    Attributes represent columns that contain actual data
    =====================================================

    The Attribute class inherits all the behaviour of L{intermine314.model.Field}
    """

    @property
    def fieldtype(self) -> str:
        return "attribute"


class Reference(Field):
    """
    References represent columns that refer to records in other tables
    ==================================================================

    In addition the the behaviour and properties of Field, references
    may also have a reverse reference, if the other record points
    back to this one as well. And all references will have their
    type upgraded to a type_class during parsing
    """

    def __init__(
        self,
        name: str,
        type_name: str,
        class_origin: "Class",
        reverse_ref: Optional[str] = None,
    ) -> None:
        """
        Constructor
        ===========

        In addition to the a parameters of Field, Reference also
        takes an optional reverse reference name (str)

        @param name: The name of the reference
        @param type_name: The name of the model.Class this refers to
        @param class_origin: The model.Class this was declared in
        @param reverse_ref: The name of the reverse reference (default: None)

        """
        self.reverse_reference_name: Optional[str] = reverse_ref
        super(Reference, self).__init__(name, type_name, class_origin)
        self.reverse_reference: Optional["Reference"] = None

    def __repr__(self) -> str:
        """
        Return a string representation
        ==============================

        @rtype: str
        """
        s = super(Reference, self).__repr__()
        if self.reverse_reference is None:
            return s
        else:
            return s + ", which links back to this as " + self.reverse_reference.name

    @property
    def fieldtype(self) -> str:
        return "reference"


class Collection(Reference):
    """
    Collections are references which refer to groups of objects
    ===========================================================

    Collections have all the same behaviour and properties as References
    """

    def __repr__(self) -> str:
        """Return a string representation"""
        ret = super(Collection, self).__repr__().replace(" is a ", " is a group of ")
        if self.reverse_reference is None:
            return ret + " objects"
        else:
            return ret.replace(", which links", " objects, which link")

    @property
    def fieldtype(self) -> str:
        return "collection"


class Class:
    """
    An abstraction of database tables in the data model
    ===================================================

    These objects refer to the table objects in the
    InterMine ORM layer.

    SYNOPSIS
    --------

    >>>  service = Service("https://www.flymine.org/query/service")
    >>>  model = service.model
    >>>
    >>>  if "Gene" in model.classes:
    ...      gene_cd = model.get_class("Gene")
    ...      print("Gene has", len(gene_cd.fields), "fields")
    ...      for field in gene_cd.fields:
    ...          print(" - ", field.name)

    OVERVIEW
    --------

    Each class can have attributes (columns) of various types,
    and can have references to other classes (tables), on either
    a one-to-one (references) or one-to-many (collections) basis

    Classes should not be instantiated by hand, but rather used
    as part of the model they belong to.

    """

    def __init__(self, name: str, parents: List[str], model: "Model", interface: bool = True) -> None:
        """
        Constructor - Creates a new Class descriptor
        ============================================

        >>> cd = intermine314.model.Class("Gene", ["SequenceFeature"])
        <intermine314.model.Class: Gene>

        This constructor is called when deserialising the
        model - you should have no need to create Classes by hand

        @param name: The name of this class
        @param parents: a list of parental names

        """
        self.name: str = name
        self.parents: List[str] = parents
        self.model: "Model" = model
        self.parent_classes: List["Class"] = []
        self.is_interface: bool = interface
        self.field_dict: Dict[str, Field] = {}
        self.has_id: bool = "Object" not in parents
        if self.has_id:
            # All InterMineObject classes have an id attribute.
            id_field = Attribute("id", "Integer", self)
            self.field_dict["id"] = id_field

    def __repr__(self) -> str:
        return "<%s.%s %s.%s>" % (
            self.__module__,
            self.__class__.__name__,
            self.model.package_name if hasattr(self.model, "package_name") else "__test__",
            self.name,
        )

    @property
    def fields(self) -> List[Field]:
        """
        The fields of this class
        ========================

        The fields are returned sorted by name. Fields
        includes all Attributes, References and Collections

        @rtype: list(L{Field})
        """
        return sorted(list(self.field_dict.values()), key=lambda field: field.name)

    def __iter__(self) -> Iterator[Field]:
        for f in list(self.field_dict.values()):
            yield f

    def __contains__(self, item: object) -> bool:
        if isinstance(item, Field):
            return item in list(self.field_dict.values())
        else:
            return str(item) in self.field_dict

    @property
    def attributes(self) -> List[Attribute]:
        """
        The fields of this class which contain data
        ===========================================

        @rtype: list(L{Attribute})
        """
        return [x for x in self.fields if isinstance(x, Attribute)]

    @property
    def references(self) -> List[Reference]:
        """
        fields which reference other objects
        ====================================

        @rtype: list(L{Reference})
        """

        def isRef(x):
            return isinstance(x, Reference) and not isinstance(x, Collection)

        return list(filter(isRef, self.fields))

    @property
    def collections(self) -> List[Collection]:
        """
        fields which reference many other objects
        =========================================

        @rtype: list(L{Collection})
        """
        return [x for x in self.fields if isinstance(x, Collection)]

    def get_field(self, name: str) -> Field:
        """
        Get a field by name
        ===================

        The standard way of retrieving a field

        @raise ModelError: if the Class does not have such a field

        @rtype: subclass of L{intermine314.model.Field}
        """
        if name in self.field_dict:
            return self.field_dict[name]
        else:
            raise ModelError("There is no field called %s in %s" % (name, self.name))

    def isa(self, other: Union["Class", str]) -> bool:
        """
        Check if self is, or inherits from other
        ========================================

        This method validates statements about inheritance.
        Returns true if the "other" is, or is within the
        ancestry of, this class

        Other can be passed as a name (str), or as the class object itself

        @rtype: boolean
        """
        if isinstance(other, Class):
            other_name = other.name
        else:
            other_name = other
        if self.name == other_name:
            return True
        if other_name in self.parents:
            return True
        for p in self.parent_classes:
            if p.isa(other):
                return True
        return False


class ComposedClass(Class):
    """
    An abstraction of dynamic objects that are in two classes
    ==========================================================

    These objects are structural unions of two or more different data-types.
    """

    def __init__(self, parts, model):
        self.is_interface = True
        self.parts = parts
        self.model = weakref.proxy(model)

    @property
    def parents(self):
        return reduce(lambda ps, cls: ps + cls.parents, self.parts, [])

    @property
    def name(self):
        return "_".join(c.name for c in self.parts)

    @property
    def has_id(self):
        return "Object" not in self.parents

    @property
    def field_dict(self):
        """The combined field dictionary of all parts"""
        fields = {}
        if self.has_id:
            # All InterMineObject classes have an id attribute.
            fields["id"] = Attribute("id", "Integer", self)
        for p in self.parts:
            fields.update(p.field_dict)
        return fields

    @property
    def parent_classes(self):
        """The flattened list of parent classes, with the parts"""
        for p in self.parts:
            all_parents = [pc for pc in p.parent_classes]
        return all_parents + self.parts


class Path(object):
    """
    A class representing a validated dotted string path
    ===================================================

    A path represents a connection between records and fields

    SYNOPSIS
    --------

        >>> service = Service("https://www.flymine.org/query/service")
            model = service.model
            path = model.make_path("Gene.organism.name")
            path.is_attribute()
        ... True
        >>> path2 = model.make_path("Gene.proteins")
            path2.is_attribute()
        ... False
        >>> path2.is_reference()
        ... True
        >>> path2.get_class()
        ... <intermine314.model.Class: gene>

    OVERVIEW
    --------

    This class is used for performing validation on dotted path strings.
    The simple act of parsing it into existence will validate the path
    to some extent, but there are additional methods for verifying certain
    relationships as well
    """

    def __init__(self, path, model, subclasses={}):
        """
        Constructor
        ===========

          >>> path = Path("Gene.name", model)

        You will not need to use this constructor directly. Instead,
        use the "make_path" method on the model to construct paths for you.

        @param path: the dotted path string (eg: Gene.proteins.name)
        @type path: str
        @param model: the model to validate the path against
        @type model: L{Model}
        @param subclasses: a dict which maps
                           subclasses (defaults to an empty dict)
        @type subclasses: dict
        """
        self.model = weakref.proxy(model)
        self.subclasses = subclasses
        if isinstance(path, Class):
            self._string = path.name
            self.parts = [path]
        else:
            self._string = str(path)
            self.parts = model.parse_path_string(str(path), subclasses)

    def __str__(self):
        return self._string

    def __repr__(self):
        return "<" + self.__module__ + "." + self.__class__.__name__ + ": " + self._string + ">"

    def prefix(self):
        """
        The path one step above this path.
        ==================================

          >>> p1 = Path("Gene.exons.name", model)
          >>> p2 = p1.prefix()
          >>> print(p2)
          ... Gene.exons

        """
        parts = list(self.parts)
        parts.pop()
        if len(parts) < 1:
            raise PathParseError(str(self) + " does not have a prefix")
        s = ".".join([x.name for x in parts])
        return Path(s, self.model._unproxied(), self.subclasses)

    def append(self, *elements):
        """
        Construct a new path by adding elements to the end of this one.
        ===============================================================

          >>> p1 = Path("Gene.exons", model)
          >>> p2 = p1.append("name")
          >>> print(p2)
          ... Gene.exons.name

        This is the inverse of prefix.
        """
        s = str(self) + "." + ".".join(elements)
        return Path(s, self.model._unproxied(), self.subclasses)

    @property
    def root(self):
        """
        The descriptor for the first part of the string.
        This should always a class descriptor.

        @rtype: L{intermine314.model.Class}
        """
        return self.parts[0]

    @property
    def end(self):
        """
        The descriptor for the last part of the string.

        @rtype: L{model.Class} or L{model.Field}
        """
        return self.parts[-1]

    def get_class(self):
        """
        Return the class object for this path, if it refers to a class
        or a reference. Attribute paths return None

        @rtype: L{model.Class}
        """
        if self.is_class():
            return self.end
        elif self.is_reference():
            if str(self) in self.subclasses:
                return self.model.get_class(self.subclasses[str(self)])
            return self.end.type_class
        else:
            return None

    end_class = property(get_class)

    def is_reference(self):
        """
        Return true if the path is a reference,
        eg: Gene.organism or Gene.proteins
        Note: Collections are ALSO references

        @rtype: boolean
        """
        return isinstance(self.end, Reference)

    def is_class(self):
        """
        Return true if the path just refers to a class, eg: Gene

        @rtype: boolean
        """
        return isinstance(self.end, Class)

    def is_attribute(self):
        """
        Return true if the path refers to an attribute, eg: Gene.length

        @rtype: boolean
        """
        return isinstance(self.end, Attribute)

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        i = hash(str(self))
        return reduce(lambda a, b: a ^ b, [hash(k) ^ hash(v) for k, v in list(self.subclasses.items())], i)


class ConstraintTree(object):
    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def __and__(self, other):
        return ConstraintTree("AND", self, other)

    def __or__(self, other):
        return ConstraintTree("OR", self, other)

    def __iter__(self):
        for n in [self.left, self.right]:
            for subn in n:
                yield subn

    def as_logic(self, codes=None, start="A"):
        if codes is None:
            codes = (chr(c) for c in range(ord(start), ord("Z")))
        return "(%s %s %s)" % (self.left.as_logic(codes), self.op, self.right.as_logic(codes))


class ConstraintNode(ConstraintTree):
    def __init__(self, *args, **kwargs):
        self.vargs = args
        self.kwargs = kwargs

    def __iter__(self):
        yield self

    def as_logic(self, codes=None, start="A"):
        if codes is None:
            codes = (chr(c) for c in range(ord(start), ord("Z")))
        return next(codes)


class CodelessNode(ConstraintNode):
    def as_logic(self, code=None, start="A"):
        return ""


class Column(object):
    """
    A representation of a path in a query that can be constrained
    =============================================================

    Column objects allow constraints to be constructed in something
    close to a declarative style
    """

    def __init__(self, path, model, subclasses={}, query=None, parent=None):
        self._model = model
        self._query = query
        self._subclasses = subclasses
        self._parent = parent
        self.filter = self.where  # alias
        if isinstance(path, Path):
            self._path = path
        else:
            self._path = model.make_path(path, subclasses)
        self._branches = {}

    def select(self, *cols):
        """
        Create a new query with this column as the base class,
        selecting the given fields.

        If no fields are given, then just this column will be selected.
        """
        q = self._model.service.new_query(str(self))
        if len(cols):
            q.select(*cols)
        else:
            q.select(self)
        return q

    def where(self, *args, **kwargs):
        """
        Create a new query based on this column,
        filtered with the given constraint.

        also available as "filter"
        """
        q = self.select()
        return q.where(*args, **kwargs)

    def __len__(self):
        """
        Return the number of values in this column.
        """
        return self.select().count()

    def __iter__(self):
        """
        Iterate over the things this column represents.

        In the case of an attribute column, that is the values it may have.
        In the caseof a reference or class column,
        it is the objects that this path may refer to.
        """
        q = self.select()
        if self._path.is_attribute():
            for row in q.rows():
                yield row[0]
        else:
            for obj in q:
                yield obj

    def __getattr__(self, name):
        if name in self._branches:
            return self._branches[name]
        cld = self._path.get_class()
        if cld is not None:
            try:
                fld = cld.get_field(name)
                branch = Column(str(self) + "." + name, self._model, self._subclasses, self._query, self)
                self._branches[name] = branch
                return branch
            except ModelError as e:
                raise AttributeError(str(e))
        raise AttributeError("No attribute '" + name + "'")

    def __str__(self):
        return str(self._path)

    def __mod__(self, other):
        if isinstance(other, tuple):
            return ConstraintNode(str(self), "LOOKUP", *other)
        else:
            return ConstraintNode(str(self), "LOOKUP", str(other))

    def __rshift__(self, other):
        return CodelessNode(str(self), str(other))

    __lshift__ = __rshift__

    def __eq__(self, other):
        if other is None:
            return ConstraintNode(str(self), "IS NULL")
        elif isinstance(other, Column):
            return ConstraintNode(str(self), "IS", str(other))
        elif hasattr(other, "make_list_constraint"):
            return other.make_list_constraint(str(self), "IN")
        elif isinstance(other, list):
            return ConstraintNode(str(self), "ONE OF", other)
        else:
            return ConstraintNode(str(self), "=", other)

    def __ne__(self, other):
        if other is None:
            return ConstraintNode(str(self), "IS NOT NULL")
        elif isinstance(other, Column):
            return ConstraintNode(str(self), "IS NOT", str(other))
        elif hasattr(other, "make_list_constraint"):
            return other.make_list_constraint(str(self), "NOT IN")
        elif isinstance(other, list):
            return ConstraintNode(str(self), "NONE OF", other)
        else:
            return ConstraintNode(str(self), "!=", other)

    def __xor__(self, other):
        if hasattr(other, "make_list_constraint"):
            return other.make_list_constraint(str(self), "NOT IN")
        elif isinstance(other, list):
            return ConstraintNode(str(self), "NONE OF", other)
        raise TypeError("Invalid argument for xor: %r" % other)

    def in_(self, other):
        if hasattr(other, "make_list_constraint"):
            return other.make_list_constraint(str(self), "IN")
        elif isinstance(other, list):
            return ConstraintNode(str(self), "ONE OF", other)
        raise TypeError("Invalid argument for in_: %r" % other)

    def __lt__(self, other):
        if isinstance(other, Column):
            self._parent._subclasses[str(self)] = str(other)
            self._parent._branches = {}
            return CodelessNode(str(self), str(other))
        try:
            return self.in_(other)
        except TypeError:
            return ConstraintNode(str(self), "<", other)

    def __le__(self, other):
        if isinstance(other, Column):
            return CodelessNode(str(self), str(other))
        try:
            return self.in_(other)
        except TypeError:
            return ConstraintNode(str(self), "<=", other)

    def __gt__(self, other):
        return ConstraintNode(str(self), ">", other)

    def __ge__(self, other):
        return ConstraintNode(str(self), ">=", other)


class Model:
    """
    A class for representing the data model of an InterMine datawarehouse
    =====================================================================

    An abstraction of the database schema

    SYNOPSIS
    --------

        >>> service = Service("https://www.flymine.org/query/service")
        >>> model = service.model
        >>> model.get_class("Gene")
        <intermine314.model.Class: Gene>

    OVERVIEW
    --------

    This class represents the data model  - ie. an abstraction
    of the database schema. It can be used to introspect what
    data is available and how it is inter-related
    """

    NUMERIC_TYPES = frozenset(
        ["int", "Integer", "float", "Float", "double", "Double", "long", "Long", "short", "Short"]
    )

    LOG = logging.getLogger("Model")

    def __init__(self, source: Any, service: Optional[Any] = None) -> None:
        """
        Constructor
        ===========

          >>> model = Model(xml)

        You will most like not need to create a model directly,
        instead get one from the Service object:

        @see: L{intermine314.webservice.Service}

        @param source: the model.xml, as a local file, string, or url
        """
        assert source is not None
        self.source: Any = source

        if service is not None:
            self.service: Optional[Any] = weakref.proxy(service)
        else:
            self.service = None

        self.classes: Dict[str, Class] = {}
        self.parse_model(source)
        self.vivify()

        # Make sugary aliases
        self.table = self.column

    def parse_model(self, source: Any) -> None:
        """
        Create classes, attributes, references and
        collections from the model.xml
        =====================================================================

        The xml can be provided as a file, url or string. This method
        is called during instantiation - it does not need to be called
        directly.

        @param source:  the model.xml, as a local file, string, or url
        @raise ModelParseError: if there is a problem parsing the source
        """
        io = None
        src = None
        try:
            io = openAnything(source)
            src = io.read()
            # Handle binary and text streams equally.
            if isinstance(src, bytes):
                src = src.decode("utf-8")
            self.LOG.debug("model = [{0}]".format(src))
            root = ET.fromstring(src)
            if root.tag == "model":
                model_node = root
            else:
                model_node = root.find("model")
            if model_node is None:
                raise ModelParseError("Error parsing model", src, "No model element found")

            self.name = model_node.get("name", "")
            self.package_name = model_node.get("package", "")
            error = "No model name or package name"
            assert self.name and self.package_name, error

            def strip_java_prefix(x):
                return re.sub(r".*\.", "", x)

            for c in model_node.findall("class"):
                class_name = c.get("name", "")
                assert class_name, "Name not defined in class"

                parents = [strip_java_prefix(p) for p in c.get("extends", "").split(" ") if len(p)]
                interface = c.get("is-interface", "") == "true"
                cl = Class(class_name, parents, self, interface)
                self.LOG.debug("Created {0}".format(cl.name))
                for a in c.findall("attribute"):
                    name = a.get("name", "")
                    type_name = strip_java_prefix(a.get("type", ""))
                    at = Attribute(name, type_name, cl)
                    cl.field_dict[name] = at
                    self.LOG.debug("set {0}.{1}".format(cl.name, at.name))
                for r in c.findall("reference"):
                    name = r.get("name", "")
                    type_name = r.get("referenced-type", "")
                    linked_field_name = r.get("reverse-reference") or ""
                    ref = Reference(name, type_name, cl, linked_field_name)
                    cl.field_dict[name] = ref
                    self.LOG.debug("set {0}.{1}".format(cl.name, ref.name))
                for co in c.findall("collection"):
                    name = co.get("name", "")
                    type_name = co.get("referenced-type", "")
                    linked_field_name = co.get("reverse-reference") or ""
                    col = Collection(name, type_name, cl, linked_field_name)
                    cl.field_dict[name] = col
                    self.LOG.debug("set {0}.{1}".format(cl.name, col.name))
                self.classes[class_name] = cl
        except Exception as error:
            model_src = src if src is not None else source
            raise ModelParseError("Error parsing model", model_src, error)
        finally:
            if io is not None:
                io.close()

    def vivify(self):
        """
        Make names point to instances and insert inherited fields
        =========================================================

        This method ensures the model is internally consistent. This method
        is called during instantiaton. It does not need to be called
        directly.

        @raise ModelError: if the names point to non-existent objects
        """
        for c in list(self.classes.values()):
            c.parent_classes = self.to_ancestry(c)
            self.LOG.debug("{0.name} < {0.parent_classes}".format(c))
            for pc in c.parent_classes:
                c.field_dict.update(pc.field_dict)
            for f in c.fields:
                f.type_class = self.classes.get(f.type_name)
                if hasattr(f, "reverse_reference_name") and f.reverse_reference_name != "":
                    rrn = f.reverse_reference_name
                    f.reverse_reference = f.type_class.field_dict[rrn]

    def to_ancestry(self, cd: Class) -> List[Class]:
        """
        Returns the lineage of the class
        ================================

            >>> classes = Model.to_ancestry(cd)

        Returns the class' parents, and all the class' parents' parents

        @rtype: list(L{intermine314.model.Class})
        """
        parents = cd.parents
        self.LOG.debug("{0} < {1}".format(cd.name, cd.parents))

        def defined(x):
            return x is not None  # weeds out the java classes

        def to_class(x):
            return self.classes.get(x)

        ancestry = list(filter(defined, list(map(to_class, parents))))
        for ancestor in ancestry:
            self.LOG.debug("{0} is ancestor of {1}".format(ancestor, cd.name))
            ancestry.extend(self.to_ancestry(ancestor))
        return ancestry

    def to_classes(self, classnames: Iterable[str]) -> List[Class]:
        """
        take a list of class names and return a list of classes
        =======================================================

            >>> classes = model.to_classes(["Gene", "Protein", "Organism"])

        This simply maps from a list of strings to a list of
        classes in the calling model.

        @raise ModelError: if the list of class names
                           includes ones that don't exist

        @rtype: list(L{intermine314.model.Class})
        """
        return list(map(self.get_class, classnames))

    def column(self, path: str, *rest: Any) -> Column:
        return Column(path, self, *rest)

    def __getattr__(self, name: str) -> Column:
        return self.column(name)

    def get_class(self, name: str) -> Class:
        """
        Get a class by its name, or by a dotted path
        ============================================

            >>> model = Model("https://www.flymine.org/query/service/model")
            >>> model.get_class("Gene")
            <intermine314.model.Class: Gene>
            >>> model.get_class("Gene.proteins")
            <intermine314.model.Class: Protein>

        This is the recommended way of retrieving a class from
        the model. As well as handling class names, you can also
        pass in a path such as "Gene.proteins" and get the
        corresponding class back (<intermine314.model.Class: Protein>)

        @raise ModelError: if the class name refers to a non-existant object

        @rtype: L{intermine314.model.Class}
        """
        if name.find(",") != -1:
            names = name.split(",")
            classes = [self.get_class(n) for n in names]
            return ComposedClass(classes, self)
        elif name.find(".") != -1:
            path = self.make_path(name)
            if path.is_attribute():
                raise ModelError("'" + str(path) + "' is not a class")
            else:
                return path.get_class()
        elif name in self.classes:
            return self.classes[name]
        else:
            raise ModelError("'" + name + "' is not a class in this model")

    def make_path(self, path: str, subclasses: Dict[str, str] = {}) -> "Path":
        """
        Return a path object for the given path string
        ==============================================

            >>> path = model.make_path("Gene.organism.name")
            <intermine314.model.Path: Gene.organism.name>

        This is recommended manner of constructing path objects.

        @type path: str
        @type subclasses: dict

        @raise PathParseError: if there is a problem parsing the path string

        @rtype: L{intermine314.model.Path}
        """
        return Path(path, self, subclasses)

    def validate_path(self, path_string, subclasses={}):
        """
        Validate a path
        ===============

            >>> try:
            ...     model.validate_path("Gene.symbol")
            ...     return "path is valid"
            ... except PathParseError:
            ...     return "path is invalid"
            "path is valid"

        When you don't need to interrogate relationships
        between paths, simply using this method to validate
        a path string is enough. It guarantees that there
        is a descriptor for each section of the string,
        with the appropriate relationships

        @raise PathParseError: if there is a problem parsing the path string
        """
        try:
            self.parse_path_string(path_string, subclasses)
            return True
        except PathParseError as e:
            raise PathParseError("Error parsing '%s' (subclasses: %s)" % (path_string, str(subclasses)), e)

    def parse_path_string(self, path_string, subclasses={}):
        """
        Parse a path string into a list of descriptors - one for each section
        =====================================================================

            >>> parts = Model.parse_path_string(string)

        This method is used when making paths from a model, and
        when validating path strings. It probably won't need to
        be called directly.

        @see: L{intermine314.model.Model.make_path}
        @see: L{intermine314.model.Model.validate_path}
        @see: L{intermine314.model.Path}
        """
        descriptors = []
        names = path_string.split(".")
        root_name = names.pop(0)

        root_descriptor = self.get_class(root_name)
        descriptors.append(root_descriptor)

        if root_name in subclasses:
            current_class = self.get_class(subclasses[root_name])
        else:
            current_class = root_descriptor

        for field_name in names:
            field = current_class.get_field(field_name)
            descriptors.append(field)

            if isinstance(field, Reference):
                key = ".".join([x.name for x in descriptors])
                if key in subclasses:
                    current_class = self.get_class(subclasses[key])
                else:
                    current_class = field.type_class
            else:
                current_class = None

        return descriptors

    def _unproxied(self):
        return self


class ModelError(ReadableException):
    pass


class PathParseError(ModelError):
    pass


class ModelParseError(ModelError):
    def __init__(self, message, source, cause=None):
        self.source = source
        super(ModelParseError, self).__init__(message, cause)

    def __str__(self):
        base = repr(self.message) + ":" + repr(self.source)
        if self.cause is None:
            return base
        else:
            return base + repr(self.cause)



from __future__ import annotations

from pathlib import Path


def build_iter_batches_kwargs(
    *,
    start,
    size,
    batch_size,
    parallel,
    page_size,
    max_workers,
    ordered,
    prefetch,
    inflight_limit,
    ordered_window_pages,
    profile,
    large_query_mode,
    pagination,
    keyset_path,
    keyset_batch_size,
):
    return {
        "start": start,
        "size": size,
        "batch_size": batch_size,
        "parallel": parallel,
        "page_size": page_size,
        "max_workers": max_workers,
        "ordered": ordered,
        "prefetch": prefetch,
        "inflight_limit": inflight_limit,
        "ordered_window_pages": ordered_window_pages,
        "profile": profile,
        "large_query_mode": large_query_mode,
        "pagination": pagination,
        "keyset_path": keyset_path,
        "keyset_batch_size": keyset_batch_size,
    }


def write_single_parquet_from_parts(
    *,
    staged_dir,
    target,
    compression,
    polars_module,
    duckdb_module,
    duckdb_quote,
):
    staged_dir = Path(staged_dir)
    target = Path(target)
    part_glob = str(staged_dir / "*.parquet")
    if not any(staged_dir.glob("*.parquet")):
        polars_module.DataFrame().write_parquet(str(target), compression=compression)
        return

    try:
        scan = polars_module.scan_parquet(part_glob)
        sink = getattr(scan, "sink_parquet", None)
        if callable(sink):
            sink(str(target), compression=compression)
            return
    except Exception:
        pass

    if duckdb_module is not None:
        part_glob_sql = duckdb_quote(part_glob)
        target_sql = duckdb_quote(str(target))
        con = duckdb_module.connect(database=":memory:")
        try:
            con.execute(
                f"COPY (SELECT * FROM read_parquet({part_glob_sql})) TO {target_sql} "
                f"(FORMAT PARQUET, COMPRESSION {compression.upper()})"
            )
        finally:
            con.close()
        return

    raise RuntimeError(
        "Cannot merge parquet parts without sink_parquet support or duckdb. "
        "Install duckdb or keep single_file=False."
    )



import requests

import xml.etree.ElementTree as etree

from intermine314.constants import DEFAULT_REQUEST_TIMEOUT_SECONDS
from intermine314.service_urls import service_root_from_payload
from intermine314.webservice import Registry

"""
Functions for better usage of saved queries.
================================================
Prompts the user to enter the API token and mine corresponding to the account.

example:

    >>>from intermine314 import query_manager as qm
"""


REGISTRY_INSTANCES_URL = Registry.DEFAULT_REGISTRY_URL.rstrip("/")
REQUEST_TIMEOUT_SECONDS = DEFAULT_REQUEST_TIMEOUT_SECONDS
HTTP_SESSION = requests.Session()

USER_QUERIES_PATH = "/user/queries"
SERVICE_VERSION_PATH = "/version"

NO_SAVED_QUERIES = "No saved queries"
NO_SUCH_QUERY_AVAILABLE = "No such query available"
INCORRECT_FORMAT = "Incorrect format"

mine = ""
token = ""


def _exception_message(exc, suffix):
    return f"An exception of type {type(exc).__name__} occurred.{suffix}"


def _request(method, url, **kwargs):
    kwargs.setdefault("timeout", REQUEST_TIMEOUT_SECONDS)
    response = HTTP_SESSION.request(method, url, **kwargs)
    response.raise_for_status()
    return response


def _request_json(method, url, **kwargs):
    return _request(method, url, **kwargs).json()


def _registry_instance_payload(mine_name):
    src = f"{REGISTRY_INSTANCES_URL}/{mine_name}"
    payload = _request_json("GET", src)
    instance = payload.get("instance")
    if not isinstance(instance, dict):
        raise KeyError("Registry response missing 'instance' block")
    return instance


def _service_root_from_instance(instance):
    return service_root_from_payload(instance, normalize=True)


def _resolve_service_root(mine_name):
    return _service_root_from_instance(_registry_instance_payload(mine_name))


def _user_queries_url(service_root):
    return f"{service_root}{USER_QUERIES_PATH}"


def _get_user_queries(service_root, api_token):
    return _request_json(
        "GET",
        _user_queries_url(service_root),
        params={"token": api_token},
    )


def _query_names(payload):
    queries = payload.get("queries", {})
    if isinstance(queries, dict):
        return list(queries.keys())
    return []


def _service_version(service_root, api_token):
    payload = _request_json(
        "GET",
        f"{service_root}{SERVICE_VERSION_PATH}",
        params={"token": api_token},
    )
    try:
        return int(payload)
    except Exception:
        return int(str(payload).strip())


def save_mine_and_token(m, t):
    """
    A function to access an account from a particular mine.
    ================================================
    example:

        >>>from intermine314 import query_manager as qm
        >>>qm.save_mine_and_token("flymine","<enter token>")
        <now you can access account linked to the token>

    """
    global mine
    global token
    mine = m
    token = t

    # test shortcut
    if mine == "mock":
        return None

    try:
        service_root = _resolve_service_root(mine)
    except Exception as exc:
        return _exception_message(exc, " Check mine")

    try:
        _ = _query_names(_get_user_queries(service_root, token))
    except Exception as exc:
        return _exception_message(exc, " Check token")
    return None


def get_all_query_names():
    """
    A function to list all the queries that are saved in a user account.
    ================================================
    example:

        >>>from intermine314 import query_manager as qm
        >>>qm.save_mine_and_token("flymine","<enter token>")
        >>>qm.get_all_query_names()
        <returns the names of all the saved queries in user account>

    """
    if mine == "mock":
        names = ["query1"]
    else:
        service_root = _resolve_service_root(mine)
        names = _query_names(_get_user_queries(service_root, token))

    if not names:
        return NO_SAVED_QUERIES
    return ", ".join(names)


def get_query(name):
    """
    A function that returns the columns that a given query constitutes.
    ================================================
    example:

        >>>from intermine314 import query_manager as qm
        >>>qm.save_mine_and_token("flymine","<enter token>")
        >>>qm.get_query('queryName')
        <returns information about the query whose name is 'queryName'>

    """
    if mine == "mock":
        ans = "c1, c2" if name == "query1" else "<saved-queries></saved-queries>"
    else:
        service_root = _resolve_service_root(mine)
        ans = _request(
            "GET",
            _user_queries_url(service_root),
            params={"filter": name, "format": "xml", "token": token},
        ).text
    if ans == "<saved-queries></saved-queries>":
        return NO_SUCH_QUERY_AVAILABLE
    return ans


def delete_query(name):
    """
    A function that deletes a given query.
    ================================================
    example:

        >>>from intermine314 import query_manager as qm
        >>>qm.save_mine_and_token("flymine","<enter token>")
        >>>qm.delete_query('queryName')
        <deletes the query whose name is 'queryName' from user's account>

    """
    if mine == "mock":
        names = ["query1", "query2"]
        if name not in names:
            return NO_SUCH_QUERY_AVAILABLE
        return name + " is deleted"

    service_root = _resolve_service_root(mine)
    names = _query_names(_get_user_queries(service_root, token))
    if name not in names:
        return NO_SUCH_QUERY_AVAILABLE

    _request(
        "DELETE",
        f"{_user_queries_url(service_root)}/{name}",
        params={"token": token},
    )
    return name + " is deleted"


def post_query(value):
    """
    A function to post a query (string containing xml or json) to a user account.
    ================================================
    example:
        >>>from intermine314 import query_manager as qm
        >>>qm.save_mine_and_token("flymine","<enter token>")
        >>>qm.post_query('<query name="" model="genomic" view="Gene.length\
            Gene.symbol" longDescription="" sortOrder="Gene.length asc">\
            </query>')
    Note that the name should be defined first.
    """
    param_name = "xml"
    root = etree.fromstring(value)
    query_name = root.attrib["name"]

    if mine == "mock":
        names = ["query1", "query2"]
        service_root = None
    else:
        service_root = _resolve_service_root(mine)
        version = _service_version(service_root, token)
        if version >= 27:
            param_name = "query"
        names = _query_names(_get_user_queries(service_root, token))

    count = 0
    for existing_name in names:
        if existing_name == query_name:
            count += 1
            print("The query name exists")
            resp = input("Do you want to replace the old query? [y/n]")
            if resp == "y":
                count = 0

    if count == 0:
        if mine == "mock":
            names = ["query1", "query2", "query3"]
        else:
            _request(
                "PUT",
                _user_queries_url(service_root),
                params={param_name: value, "token": token},
            )
            names = _query_names(_get_user_queries(service_root, token))

        if query_name not in names:
            print("Note: name should contain no special symbol and should be defined first")
            return INCORRECT_FORMAT
        return query_name + " is posted"

    print("Use a query name other than " + query_name)



from functools import wraps
from intermine314.errors import ServiceError


def requires_version(required):
    error_fmt = "Service must be at version %s, but is at %s"

    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if self.version < required:
                raise ServiceError(error_fmt % (required, self.version))
            return f(self, *args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

import re
from pathlib import Path

from intermine314.constants import DEFAULT_PARALLEL_PAGE_SIZE, PRODUCTION_WORKFLOW_ELT, PRODUCTION_WORKFLOW_ETL
from intermine314.mine_registry import resolve_production_plan
from intermine314.optional_deps import require_duckdb as _require_duckdb
from intermine314.webservice import Service

_DUCKDB_IDENTIFIER_PATTERN = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")


def _build_query(
    *,
    mine_url: str,
    root_class: str,
    views: list[str],
    joins: list[str],
):
    service = Service(mine_url)
    query = service.new_query(root_class)
    query.add_view(*views)
    for join in joins:
        query.add_join(join, "OUTER")
    return query


def _effective_parallel_args(
    *,
    mine_url: str,
    size: int | None,
    workflow: str,
    production_profile: str,
    max_workers: int | None,
    ordered,
    parallel_profile: str | None,
    large_query_mode: bool | None,
    prefetch: int | None,
    inflight_limit: int | None,
):
    plan = resolve_production_plan(
        mine_url,
        size,
        workflow=workflow,
        production_profile=production_profile,
    )
    workers = int(max_workers if max_workers is not None else plan["workers"])
    return plan, {
        "max_workers": workers,
        "ordered": plan["ordered"] if ordered is None else ordered,
        "profile": str(parallel_profile or plan["parallel_profile"]),
        "large_query_mode": bool(plan["large_query_mode"] if large_query_mode is None else large_query_mode),
        "prefetch": prefetch if prefetch is not None else plan.get("prefetch"),
        "inflight_limit": inflight_limit if inflight_limit is not None else plan.get("inflight_limit"),
    }


def fetch_from_mine(
    *,
    mine_url: str,
    root_class: str,
    views: list[str],
    joins: list[str] | None = None,
    start: int = 0,
    size: int | None = None,
    page_size: int = DEFAULT_PARALLEL_PAGE_SIZE,
    workflow: str = PRODUCTION_WORKFLOW_ELT,
    production_profile: str = "auto",
    max_workers: int | None = None,
    ordered=None,
    parallel_profile: str | None = None,
    large_query_mode: bool | None = None,
    prefetch: int | None = None,
    inflight_limit: int | None = None,
    ordered_window_pages: int = 10,
    parquet_path: str | Path | None = None,
    parquet_compression: str = "zstd",
    duckdb_database: str = ":memory:",
    duckdb_table: str = "results",
):
    """
    Fetch mine data with a mine-aware production profile.

    Workflows:
    - ELT: parallel fetch -> parquet -> duckdb view
    - ETL: parallel fetch -> polars dataframe -> duckdb table
    """
    workflow = str(workflow or PRODUCTION_WORKFLOW_ELT).strip().lower()
    if workflow not in {PRODUCTION_WORKFLOW_ELT, PRODUCTION_WORKFLOW_ETL}:
        raise ValueError("workflow must be 'elt' or 'etl'")
    if not _DUCKDB_IDENTIFIER_PATTERN.fullmatch(str(duckdb_table)):
        raise ValueError("duckdb_table must be a valid SQL identifier")

    query = _build_query(
        mine_url=mine_url,
        root_class=root_class,
        views=list(views),
        joins=list(joins or []),
    )
    plan, parallel_args = _effective_parallel_args(
        mine_url=mine_url,
        size=size,
        workflow=workflow,
        production_profile=production_profile,
        max_workers=max_workers,
        ordered=ordered,
        parallel_profile=parallel_profile,
        large_query_mode=large_query_mode,
        prefetch=prefetch,
        inflight_limit=inflight_limit,
    )

    common_args = {
        "start": start,
        "size": size,
        "parallel": True,
        "page_size": page_size,
        "max_workers": parallel_args["max_workers"],
        "ordered": parallel_args["ordered"],
        "prefetch": parallel_args["prefetch"],
        "inflight_limit": parallel_args["inflight_limit"],
        "ordered_window_pages": ordered_window_pages,
        "profile": parallel_args["profile"],
        "large_query_mode": parallel_args["large_query_mode"],
        "pagination": "auto",
    }

    duckdb = _require_duckdb("fetch_from_mine()")
    con = duckdb.connect(database=duckdb_database)

    if workflow == PRODUCTION_WORKFLOW_ELT:
        if parquet_path is None:
            raise ValueError("parquet_path is required for ELT workflow")
        parquet_path = str(Path(parquet_path))
        query.to_parquet(
            parquet_path,
            compression=parquet_compression,
            single_file=True,
            **common_args,
        )
        con.execute(
            f'CREATE OR REPLACE VIEW "{duckdb_table}" AS SELECT * FROM read_parquet(?)',
            [parquet_path],
        )
        return {
            "workflow": workflow,
            "production_plan": plan,
            "parquet_path": parquet_path,
            "duckdb_table": duckdb_table,
            "duckdb_connection": con,
        }

    dataframe = query.dataframe(**common_args)
    con.register("_intermine314_polars_df", dataframe)
    con.execute(f'CREATE OR REPLACE TABLE "{duckdb_table}" AS SELECT * FROM _intermine314_polars_df')
    try:
        con.unregister("_intermine314_polars_df")
    except Exception:
        pass
    return {
        "workflow": workflow,
        "production_plan": plan,
        "dataframe": dataframe,
        "duckdb_table": duckdb_table,
        "duckdb_connection": con,
    }



from intermine314.webservice import Registry, Service

"""
Functions for making use of registry data
================================================

"""

NO_SUCH_MINE = "No such mine available"


def _safe_registry_info(mine):
    registry = Registry()
    info = registry.info(mine)
    return registry, info


def getVersion(mine):
    """
    A function to return the API version, release version and
    InterMine version numbers
    ================================================
    example:

        >>> from intermine314 import registry
        >>> registry.getVersion('flymine')
        >>> {'API Version:': '30', 'Release Version:': '48 2019 October',
        'InterMine Version:': '4.1.0'}

    """
    try:
        _, info = _safe_registry_info(mine)
        return {
            "API Version:": info.get("api_version"),
            "Release Version:": info.get("release_version"),
            "InterMine Version:": info.get("intermine_version"),
        }
    except Exception:
        return NO_SUCH_MINE


def getInfo(mine):
    """
    A function to get information about a mine
    ================================================
    example:

        >>> from intermine314 import registry
        >>> registry.getInfo('flymine')
        Description:  An integrated database for Drosophila genomics
        URL: https://www.flymine.org/flymine
        API Version: 25
        Release Version: 45.1 2017 August
        InterMine Version: 1.8.5
        Organisms:
        D. melanogaster
        Neighbours:
        MODs

    """
    try:
        _, info = _safe_registry_info(mine)
        print("Description: " + (info.get("description") or ""))
        print("URL: " + (info.get("url") or ""))
        print("API Version: " + (info.get("api_version") or ""))
        print("Release Version: " + (info.get("release_version") or ""))
        print("InterMine Version: " + (info.get("intermine_version") or ""))
        print("Organisms: ")
        for organism in info.get("organisms", []):
            print(organism)
        print("Neighbours: ")
        for neighbour in info.get("neighbours", []):
            print(neighbour)
        return None
    except Exception:
        return NO_SUCH_MINE


def getData(mine):
    """
    A function to get datasets corresponding to a mine
    ================================================
    example:

        >>> from intermine314 import registry
        >>> registry.getData('flymine')
        Name: Affymetrix array: Drosophila1
        Name: Affymetrix array: Drosophila2
        Name: Affymetrix array: GeneChip Drosophila Genome 2.0 Array
        Name: Affymetrix array: GeneChip Drosophila Genome Array
        Name: Anoph-Expr data set
        Name: BDGP cDNA clone data set.....


    """
    try:
        registry, info = _safe_registry_info(mine)
        service_root = info.get("url") or registry.service_root(mine)
        if not service_root:
            return NO_SUCH_MINE

        service = Service(service_root)
        dataset_names = []
        query_shapes = (
            ("DataSet", ("DataSet.name", "DataSet.url"), ("DataSet.name", "name")),
            ("Dataset", ("Dataset.name", "Dataset.url"), ("Dataset.name", "name")),
        )
        for class_name, views, keys in query_shapes:
            try:
                query = service.new_query(class_name)
                query.add_view(*views)
                for row in query.rows(row="dict", start=0, size=500):
                    value = None
                    for key in keys:
                        if key in row and row[key]:
                            value = row[key]
                            break
                    if value:
                        dataset_names.append(str(value))
                break
            except Exception:
                continue

        for name in sorted(set(dataset_names)):
            print("Name: " + name)
        return None
    except Exception:
        return NO_SUCH_MINE


def getMines(organism=None):
    """
    A function to get mines containing the organism
    ================================================
    example:

        >>> from intermine314 import registry
        >>> registry.getMines('D. melanogaster')
        FlyMine
        FlyMine Beta
        XenMine

    """
    try:
        mines = Service.get_all_mines(organism=organism)
    except Exception:
        return NO_SUCH_MINE

    names = sorted(set(m.get("name") for m in mines if isinstance(m, dict) and m.get("name")))
    if not names:
        return NO_SUCH_MINE
    for name in names:
        print(name)
    return None



import weakref
import logging

import codecs
from urllib.parse import urlencode

from intermine314.constants import DEFAULT_LIST_ENTRIES_BATCH_SIZE
from intermine314.results import JSONIterator, EnrichmentLine
from intermine314.model import ConstraintNode
from intermine314.errors import ServiceError


class List:
    """
    Class for representing a List on an InterMine Webservice
    ========================================================

    Lists represent stored collections of data and saved result
    sets in an InterMine data warehouse. This class is an abstraction
    of this information, and provides mechanisms for managing the
    data.

    SYNOPSIS
    --------

    example::

        >>> from intermine314.webservice import Service
        >>>
        >>> flymine = Service("www.flymine.org/query", "SOMETOKEN")
        >>> new_list = flymine.create_list(["h", "zen", "eve", "bib"], "Gene", name="My New List")
        >>>
        >>> another_list = flymine.get_list("Some other list")
        >>> combined_list = new_list | another_list # Same syntax as for sets
        >>> combined_list.name = "Union of the other lists"
        >>>
        >>> print("The combination of the two lists has %d elements" % combined_list.size)
        >>> print("The combination of the two lists has %d elements" % len(combined_list))
        >>>
        >>> for row in combined_list:
        ...     print(row)

    OVERVIEW
    --------

    Lists are created from a webservice, and can be manipulated in various ways.
    The operations are::
        * Union: this | that
        * Intersection: this & that
        * Symmetric Difference: this ^ that
        * Asymmetric Difference (subtraction): this - that
        * Appending: this += that

    Lists can be created from a list of identifiers that could be::
        * stored in a file
        * held in a list or set
        * contained in a string
    In all these cases the syntax is the same:

        >>> new_list = service.create_list(content, type, name="Some name", description="Some description", tags=["some", "tags"])

    Lists can also be created from a query's result with the exact
    same syntax. In the case of queries, the type is not required,
    but the query should have just one view, and it should be an id.

        >>> query = service.new_query()
        >>> query.add_view("Gene.id")
        >>> query.add_constraint("Gene.length", "<", 100)
        >>> new_list = service.create_list(query, name="Short Genes")

    """

    LOG = logging.getLogger("List")

    def __init__(self, **args):
        """
        Constructor
        ===========

        Do not construct these objects yourself. They should be
        fetched from a service or constructed using the "create_list"
        method.
        """
        self.LOG.debug(args)
        try:
            self._service = args["service"]
            self._manager = weakref.proxy(args["manager"])
            self._name = args["name"]
            self._title = args["title"]
            self._description = args.get("description")
            self._list_type = args["type"]
            self._size = int(args["size"])
            self._date_created = args.get("dateCreated")
            self._is_authorized = args.get("authorized")
            self._status = args.get("status")

            if self._is_authorized is None:
                self._is_authorized = True

            if "tags" in args:
                tags = args["tags"]
            else:
                tags = []

            self._tags = frozenset(tags)
        except KeyError:
            raise ValueError("Missing argument")
        self.unmatched_identifiers = set([])

    @property
    def date_created(self):
        """When this list was originally created"""
        return self._date_created

    @property
    def tags(self):
        """The tags associated with this list"""
        return self._tags

    @property
    def description(self):
        """The human readable description of this list"""
        return self._description

    @property
    def title(self):
        """The fixed title of this list"""
        return self._title

    @property
    def status(self):
        """The upgrade status of this list"""
        return self._status

    @property
    def is_authorized(self):
        """Whether or not the current user is authorised to make changes to this list"""
        return self._is_authorized

    @property
    def list_type(self):
        """The type of the InterMine objects this list can contain"""
        return self._list_type

    def get_name(self):
        """The name of the list used to access it programmatically"""
        return self._name

    def set_name(self, new_name):
        """
        Set the name of the list
        ========================

        Setting the list's name causes the list's name to be updated on the server.
        """
        if self._name == new_name:
            return
        uri = self._service.root + self._service.LIST_RENAME_PATH
        params = {"oldname": self._name, "newname": new_name}
        uri += "?" + urlencode(params)
        resp = self._service.opener.open(uri)
        data = resp.read()
        resp.close()
        new_list = self._manager.parse_list_upload_response(data)
        self._name = new_name

    def del_name(self):
        """Raises an error - lists must always have a name"""
        raise AttributeError("List names cannot be deleted, only changed")

    @property
    def size(self):
        """Return the number of elements in the list. Also available as len(obj)"""
        return self._size

    @property
    def count(self):
        """Alias for obj.size. Also available as len(obj)"""
        return self.size

    def __len__(self):
        """Returns the number of elements in the object"""
        return self.size

    name = property(get_name, set_name, del_name, "The name of this list")

    def _add_failed_matches(self, ids):
        if ids is not None:
            self.unmatched_identifiers.update(ids)

    def __str__(self):
        string = self.name + " (" + str(self.size) + " " + self.list_type + ")"
        if self.date_created:
            string += " " + self.date_created
        if self.description:
            string += " " + self.description
        return string

    def delete(self):
        """
        Delete this list from the webservice
        ====================================

        Calls the webservice to delete this list immediately. This
        object should not be used after this method is called - attempts
        to do so will raise errors.
        """
        self._manager.delete_lists([self])

    def to_query(self):
        """
        Construct a query to fetch the items in this list
        =================================================

        Return a new query constrained to the objects in this list,
        and with a single view column of the objects ids.

        @rtype: intermine314.query.Query
        """
        q = self._service.new_query(self.list_type)
        q.add_constraint(self.list_type, "IN", self.name)
        return q

    def make_list_constraint(self, path, op):
        """
        Implementation of trait that allows use of these objects in list constraints
        """
        return ConstraintNode(path, op, self.name)

    def display(self):
        """
        Display the contents present in the list
        ========================================

        Format of the output
        Row #:
        key1 = value1
        key2 = value2
        .
        .
        """
        c = 1
        for i in self:
            s = str(i)
            print("Row {}:".format(c))
            c += 1
            k = 0
            while s[k] != "(":
                k += 1
            s = s[k + 1 :]
            s = s.split(",")
            for j in s:
                print(j.strip())
            print()

    def __iter__(self):
        """Return an iterator over the objects in this list, with all attributes selected for output"""
        return iter(self.to_query())

    def get_entries(self, batch_size=DEFAULT_LIST_ENTRIES_BATCH_SIZE):
        """
        Yield list member identifiers in batches.
        """
        if batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")
        query = self._service.new_query(self.list_type)
        query.add_view(self.list_type + ".id")
        query.add_constraint(self.list_type, "IN", self.name)
        count = query.count()
        for start in range(0, count, batch_size):
            for row in query.results(row="list", start=start, size=batch_size):
                if row:
                    yield row[0]

    def __getitem__(self, index):
        """Get a member of this list by index"""
        if not isinstance(index, int):
            raise IndexError("Expected an integer key - got %s" % (index))
        if index < 0:  # handle negative indices.
            i = self.size + index
        else:
            i = index

        if i not in range(self.size):
            raise IndexError("%d is not a valid index for a list of size %d" % (index, self.size))

        return self.to_query().first(start=i, row="jsonobjects")

    def __and__(self, other):
        """
        Intersect this list and another
        """
        return self._manager.intersect([self, other])

    def __iand__(self, other):
        """
        Intersect this list and another, and replace this list with the result of the
        intersection
        """
        intersection = self._manager.intersect([self, other], description=self.description, tags=self.tags)
        self.delete()
        intersection.name = self.name
        return intersection

    def __or__(self, other):
        """
        Return the union of this list and another
        """
        return self._manager.union([self, other])

    def __add__(self, other):
        """
        Return the union of this list and another
        """
        return self._manager.union([self, other])

    def __iadd__(self, other):
        """
        Append other to this list.
        """
        return self.append(other)

    def _do_append(self, content):
        name = self.name
        data = None

        try:
            ids = codecs.open(content, "r", "UTF-8").read()
        except (TypeError, IOError):
            if hasattr(content, "strip") and hasattr(content, "encode"):
                ids = content  # probably a string.
            else:
                try:
                    ids = "\n".join(map(lambda x: '"' + x + '"', iter(content)))
                except TypeError:
                    content = self._manager._get_listable_query(content)
                    uri = content.get_list_append_uri()
                    params = content.to_query_params()
                    params["listName"] = name
                    params["path"] = None
                    form = urlencode(params)
                    resp = self._service.opener.open(uri, form)
                    data = resp.read()

        if data is None:
            uri = self._service.root + self._service.LIST_APPENDING_PATH
            query_form = {"name": name}
            uri += "?" + urlencode(query_form)
            data = self._service.opener.post_plain_text(uri, ids)

        new_list = self._manager.parse_list_upload_response(data)
        self.unmatched_identifiers.update(new_list.unmatched_identifiers)
        self._size = new_list.size
        return self

    def append(self, appendix):
        """Append the arguments to this list"""
        try:
            return self._do_append(self._manager.union(appendix))
        except Exception:
            return self._do_append(appendix)

    def calculate_enrichment(self, widget, background=None, correction="Holm-Bonferroni", maxp=0.05, filter=""):
        """
        Perform an enrichment calculation on this list
        ==============================================

        example::

            >>> for item in service.get_list("some list").calculate_enrichment("thingy_enrichment"):
            ...     print(item.identifier, item.p_value)

        Gets an iterator over the rows for an enrichment calculation. Each row represents
        a record with the following properties:
            * identifier {str}
            * p-value {float}
            * matches {int}
            * description {str}

        The enrichment row object may be treated as an object with property access, or as
        a dictionary, supporting key lookup with the [] operator:

            >>> p_value = row['p-value']

        """
        if self._service.version < 8:
            raise ServiceError("This service does not support enrichment requests")
        params = dict(list=self.name, widget=widget, correction=correction, maxp=maxp, filter=filter)
        if background is not None:
            if self._service.version < 11:
                raise ServiceError("This service does not support custom background populations")
            params["population"] = background
        form = urlencode(params)
        uri = self._service.root + self._service.LIST_ENRICHMENT_PATH
        resp = self._service.opener.open(uri, form)
        return JSONIterator(resp, EnrichmentLine)

    def __xor__(self, other):
        """Calculate the symmetric difference of this list and another"""
        return self._manager.xor([self, other])

    def __ixor__(self, other):
        """Calculate the symmetric difference of this list and another and replace this list with the result"""
        diff = self._manager.xor([self, other], description=self.description, tags=self.tags)
        self.delete()
        diff.name = self.name
        return diff

    def __sub__(self, other):
        """Subtract the other from this list"""
        return self._manager.subtract([self], [other])

    def __isub__(self, other):
        """Replace this list with the subtraction of the other from this list"""
        subtr = self._manager.subtract([self], [other], description=self.description, tags=self.tags)
        self.delete()
        subtr.name = self.name
        return subtr

    def add_tags(self, *tags):
        """
        Tag this list with one or more categories
        =========================================

        Calls the server to add these tags, and updates this lists tags.
        """
        self._tags = frozenset(self._manager.add_tags(self, tags))

    def remove_tags(self, *tags):
        """
        Remove tags associated with this list.
        ======================================

        Calls the server to remove these tags, and updates this lists tags.
        """
        self._tags = frozenset(self._manager.remove_tags(self, tags))

    def update_tags(self, *tags):
        """
        Remove tags associated with this list.
        ======================================

        Calls the server to remove these tags, and updates this lists tags.
        """
        self._tags = frozenset(self._manager.get_tags(self))



from xml.dom import minidom
from contextlib import closing
import json
from uuid import uuid4

import requests

from urllib.parse import urlencode
from collections.abc import MutableMapping as DictMixin

# Local intermine314 imports
from intermine314.model import Model, Attribute, Reference, Collection, Column
from intermine314.lists.listmanager import ListManager
from intermine314.errors import ServiceError, WebserviceError
from intermine314.results import InterMineURLOpener, ResultIterator
from intermine314 import idresolution
from intermine314.decorators import requires_version
from intermine314.constants import DEFAULT_LIST_CHUNK_SIZE, DEFAULT_REQUEST_TIMEOUT_SECONDS
from intermine314.service_urls import normalize_service_root, service_root_from_payload

"""
Webservice Interaction Routines for InterMine Webservices
=========================================================

Classes for dealing with communication with an InterMine
RESTful webservice.

"""

__author__ = "Alex Kalderimis"
__organization__ = "InterMine"
__license__ = "LGPL"
__contact__ = "toffe.kari@gmail.com"

_QUERY_CLASS = None
_TEMPLATE_CLASS = None


def _query_classes():
    global _QUERY_CLASS, _TEMPLATE_CLASS
    if _QUERY_CLASS is None or _TEMPLATE_CLASS is None:
        from intermine314.query import Query, Template

        _QUERY_CLASS = Query
        _TEMPLATE_CLASS = Template
    return _QUERY_CLASS, _TEMPLATE_CLASS


class Registry(DictMixin):
    """
    A Class representing an InterMine registry.
    ===========================================

    Registries are web-services that mines can automatically register
    themselves with, and thus enable service discovery by clients.

    SYNOPSIS
    --------

    example::

        from intermine314.webservice import Registry

        # Connect to the default registry service
        # at www.intermine.org/registry
        registry = Registry()

        # Find all the available mines:
        for name, mine in registry.items():
            print(name, mine.version)

        # Dict-like interface for accessing mines.
        flymine = registry["flymine"]

        # The mine object is a Service
        for gene in flymine.select("Gene.*").results():
            process(gene)

    This class is meant to aid with interoperation between
    mines by allowing them to discover one-another, and
    allow users to always have correct connection information.
    """

    MINES_PATH = "/mines.json"
    INSTANCES_PATH = "/service/instances"
    DEFAULT_REGISTRY_URL = "https://registry.intermine.org/service/instances"

    def __init__(self, registry_url=DEFAULT_REGISTRY_URL):
        self.registry_url = registry_url.rstrip("/")
        opener = InterMineURLOpener()
        data = opener.open(self._list_url()).read()
        mine_data = json.loads(ensure_str(data))
        mines = self._extract_mines(mine_data)
        self.__mine_dict = dict(((mine["name"], mine) for mine in mines))
        self.__synonyms = dict(((name.lower(), name) for name in list(self.__mine_dict.keys())))
        self.__mine_cache = {}

    def _list_url(self):
        if self.registry_url.endswith(self.MINES_PATH.rstrip("/")):
            return self.registry_url
        if self.registry_url.endswith(self.INSTANCES_PATH.rstrip("/")):
            return self.registry_url
        if self.registry_url.endswith("/registry"):
            return self.registry_url + self.MINES_PATH
        return self.registry_url + self.INSTANCES_PATH

    def _extract_mines(self, data):
        if "instances" in data:
            return data["instances"]
        if "mines" in data:
            return data["mines"]
        raise ServiceError("Registry response missing expected 'instances' or 'mines' data")

    def _service_root(self, mine):
        try:
            return service_root_from_payload(mine)
        except KeyError:
            raise KeyError("Missing service URL for mine")

    def __contains__(self, name):
        return name.lower() in self.__synonyms

    def __getitem__(self, name):
        lc = name.lower()
        if lc in self.__synonyms:
            if lc not in self.__mine_cache:
                mine = self.__mine_dict[self.__synonyms[lc]]
                self.__mine_cache[lc] = Service(self._service_root(mine))
            return self.__mine_cache[lc]
        else:
            raise KeyError("Unknown mine: " + name)

    def __setitem__(self, name, item):
        raise NotImplementedError("You cannot add items to a registry")

    def __delitem__(self, name):
        raise NotImplementedError("You cannot remove items from a registry")

    def __len__(self):
        return len(self.__mine_dict)

    def __iter__(self):
        return iter(self.__mine_dict)

    def keys(self):
        return list(self.__mine_dict.keys())

    def info(self, name):
        """
        Return the registry info dictionary for a mine.
        """
        lc = name.lower()
        if lc in self.__synonyms:
            return self.__mine_dict[self.__synonyms[lc]]
        raise KeyError("Unknown mine: " + name)

    def service_root(self, name):
        """
        Return the service root URL for a mine.
        """
        return self._service_root(self.info(name))

    def all_mines(self, organism=None):
        """
        Return a list of registry info dictionaries, optionally filtered by organism.
        """
        mines = list(self.__mine_dict.values())
        if organism is None:
            return mines
        target = organism.strip()
        filtered = []
        for mine in mines:
            organisms = mine.get("organisms") or []
            for entry in organisms:
                if entry.strip() == target:
                    filtered.append(mine)
                    break
        return filtered


def ensure_str(stringlike):
    if isinstance(stringlike, bytes):
        return stringlike.decode("utf-8")
    if isinstance(stringlike, str):
        return stringlike
    return str(stringlike)


class Service:
    """
    A class representing connections to different InterMine WebServices
    ===================================================================

    The intermine314.webservice.Service class is the main interface for the user.
    It will provide access to queries and templates, as well as doing the
    background task of fetching the data model, and actually requesting
    the query results.

    SYNOPSIS
    --------

    example::

      from intermine314.webservice import Service
      service = Service("https://www.flymine.org/query/service")

      template = service.get_template("Gene_Pathways")
      for row in template.results(A={"value":"zen"}):
        do_something_with(row)
        ...

      query = service.new_query()
      query.add_view("Gene.symbol", "Gene.pathway.name")
      query.add_constraint("Gene", "LOOKUP", "zen")
      for row in query.results():
        do_something_with(row)
        ...

      new_list = service.create_list("some/file/with.ids", "Gene")
      list_on_server = service.get_list("On server")
      in_both = new_list & list_on_server
      in_both.name = "Intersection of these lists"
      for row in in_both:
        do_something_with(row)
        ...

    OVERVIEW
    --------
    The two methods the user will be most concerned with are:
      - L{Service.new_query}: constructs a new query to query a service with
      - L{Service.get_template}: gets a template from the service
      - L{ListManager.create_list}: creates a new list on the service

    For list management information, see L{ListManager}.

    TERMINOLOGY
    -----------
    X{Query} is the term for an arbitrarily complex structured request for
    data from the webservice. The user is responsible for specifying the
    structure that determines what records are returned, and what information
    about each record is provided.

    X{Template} is the term for a predefined "Query", ie: one that has been
    written and saved on the webservice you will access. The definition
    of the query is already done, but the user may want to specify the
    values of the constraints that exist on the template. Templates are accessed
    by name, and while you can easily introspect templates, it is assumed
    you know what they do when you use them

    X{List} is a saved result set containing a set of objects previously identified
    in the database. Lists can be created and managed using this client library.

    @see: L{intermine314.query}
    """

    QUERY_PATH = "/query/results"
    LIST_ENRICHMENT_PATH = "/list/enrichment"
    WIDGETS_PATH = "/widgets"
    SEARCH_PATH = "/search"
    QUERY_LIST_UPLOAD_PATH = "/query/tolist"
    QUERY_LIST_APPEND_PATH = "/query/append/tolist"
    MODEL_PATH = "/model"
    TEMPLATES_PATH = "/templates"
    TEMPLATEQUERY_PATH = "/template/results"
    ALL_TEMPLATES_PATH = "/alltemplates"
    LIST_PATH = "/lists"
    LIST_CREATION_PATH = "/lists"
    LIST_RENAME_PATH = "/lists/rename"
    LIST_APPENDING_PATH = "/lists/append"
    LIST_TAG_PATH = "/list/tags"
    SAVEDQUERY_PATH = "/savedqueries/xml"
    VERSION_PATH = "/version/ws"
    RELEASE_PATH = "/version/release"
    SCHEME = "http://"
    SERVICE_RESOLUTION_PATH = "/check/"
    IDS_PATH = "/ids"
    USERS_PATH = "/users"
    REGISTRY_URL = Registry.DEFAULT_REGISTRY_URL

    def __init__(
        self,
        root,
        username=None,
        password=None,
        token=None,
        prefetch_depth=1,
        prefetch_id_only=False,
        request_timeout=DEFAULT_REQUEST_TIMEOUT_SECONDS,
    ):
        """
        Constructor
        ===========

        Construct a connection to a webservice::

            url = "https://www.flymine.org/query/service"

            # An unauthenticated connection - access to all public data
            service = Service(url)

            # An authenticated connection - access to private and public data
            service = Service(url, token="ABC123456")


        @param root: the root url of the webservice (required)
        @param username: your login name (optional)
        @param password: your password (required if a username is given)
        @param token: your API access token(optional - used in preference to username and password)

        @raise ServiceError: if the version cannot be fetched and parsed
        @raise ValueError:   if a username is supplied, but no password

        There are two alternative authentication systems supported by InterMine
        webservices. The first is username and password authentication, which
        is supported by all webservices. Newer webservices (version 6+)
        also support API access token authentication, which is the recommended
        system to use. Token access is more secure as you will never have
        to transmit your username or password, and the token can be easily changed
        or disabled without changing your webapp login details.

        """
        root = normalize_service_root(root)

        self.root = root
        self.prefetch_depth = prefetch_depth
        self.prefetch_id_only = prefetch_id_only
        self.request_timeout = request_timeout
        # Initialize empty cached data.
        self._templates = None
        self._all_templates = None
        self._all_templates_names = None
        self._model = None
        self._version = None
        self._release = None
        self._widgets = None
        self._list_manager = ListManager(self)
        self.__missing_method_name = None
        if token:
            if token == "random":
                token = self.get_anonymous_token(url=root)
            self.opener = InterMineURLOpener(token=token, request_timeout=self.request_timeout)
        elif username:
            if token:
                raise ValueError("Both username and token credentials supplied")

            if not password:
                raise ValueError("Username given, but no password supplied")

            self.opener = InterMineURLOpener((username, password), request_timeout=self.request_timeout)
        else:
            self.opener = InterMineURLOpener(request_timeout=self.request_timeout)

        try:
            self.version
        except WebserviceError as e:
            raise ServiceError("Could not validate service - is the root url (%s) correct? %s" % (root, e))

        if token and self.version < 6:
            raise ServiceError("This service does not support API access token authentication")

        # Set up sugary aliases
        self.query = self.new_query

    # Delegated list methods

    LIST_MANAGER_METHODS = frozenset(
        ["get_list", "get_all_lists", "get_all_list_names", "create_list", "get_list_count", "delete_lists", "l"]
    )

    def get_anonymous_token(self, url):
        """
        Generates an anonymous session token valid for 24 hours
        =======================================================
        """
        url += "/session"
        token = requests.get(url=url, timeout=self.request_timeout)
        token = token.json()["token"]
        return token

    def get_mine_info(self, mine_name, registry_url=None):
        """
        Fetch registry info for a given mine.
        """
        registry = Registry(registry_url=registry_url or self.REGISTRY_URL)
        return registry.info(mine_name)

    @classmethod
    def get_all_mines(cls, organism=None, registry_url=None):
        """
        Fetch all registry mines, optionally filtered by organism.
        """
        registry = Registry(registry_url=registry_url or cls.REGISTRY_URL)
        return registry.all_mines(organism=organism)

    def list_manager(self):
        """
        Get a new ListManager to use with this service.
        ===============================================

        This method is primarily useful as a context manager
        when creating temporary lists, since on context exit all
        temporary lists will be cleaned up::

            with service.list_manager() as manager:
                temp_a = manager.create_list(file_a, "Gene")
                temp_b = manager.create_list(file_b, "Gene")
                for gene in (temp_a & temp_b):
                    print(gene.primaryIdentifier, "is in both")

        @rtype: ListManager
        """
        return ListManager(self)

    def list_templates(self, include=None, exclude=None, limit=None):
        """
        Return template names, with optional substring include/exclude filters.
        """
        names = sorted(self.templates.keys())
        include_terms = [str(term).lower() for term in (include or []) if str(term).strip()]
        exclude_terms = [str(term).lower() for term in (exclude or []) if str(term).strip()]

        if include_terms:
            names = [name for name in names if any(term in name.lower() for term in include_terms)]
        if exclude_terms:
            names = [name for name in names if not any(term in name.lower() for term in exclude_terms)]
        if limit is not None:
            names = names[: max(0, int(limit))]
        return names

    def create_batched_lists(
        self,
        identifiers,
        list_type,
        chunk_size=DEFAULT_LIST_CHUNK_SIZE,
        name_prefix="intermine314_batch",
        description=None,
        tags=None,
    ):
        """
        Create multiple server-side lists from identifiers in fixed-size chunks.
        """
        if not isinstance(chunk_size, int) or isinstance(chunk_size, bool):
            raise TypeError("chunk_size must be an integer")
        if chunk_size <= 0:
            raise ValueError("chunk_size must be a positive integer")
        if not list_type:
            raise ValueError("list_type is required")

        tags = list(tags or [])
        run_id = uuid4().hex[:8]
        batch = []
        created = []
        batch_index = 1

        def flush(current):
            nonlocal batch_index
            if not current:
                return
            name = f"{name_prefix}_{run_id}_{batch_index:05d}"
            batch_index += 1
            created.append(
                self.create_list(
                    current,
                    list_type=list_type,
                    name=name,
                    description=description,
                    tags=tags,
                )
            )

        for identifier in identifiers:
            if identifier is None:
                continue
            value = str(identifier).strip()
            if not value:
                continue
            batch.append(value)
            if len(batch) >= chunk_size:
                flush(batch)
                batch = []

        flush(batch)
        return created

    def __getattribute__(self, name):
        return object.__getattribute__(self, name)

    def __getattr__(self, name):
        if name in self.LIST_MANAGER_METHODS:
            method = getattr(self._list_manager, name)
            return method
        raise AttributeError("Could not find " + name)

    def __del__(self):  # On going out of scope, try and clean up.
        try:
            self._list_manager.delete_temporary_lists()
        except ReferenceError:
            pass

    @property
    def version(self):
        """
        Returns the webservice version
        ==============================

        The version specifies what capabilities a
        specific webservice provides. The most current
        version is 3

        may raise ServiceError: if the version cannot be fetched

        @rtype: int
        """
        try:
            if self._version is None:
                try:
                    url = self.root + self.VERSION_PATH
                    self._version = int(self.opener.open(url).read())
                except ValueError as e:
                    raise ServiceError("Could not parse a valid webservice version: " + str(e))
        except AttributeError as e:
            raise Exception(e)
        return self._version

    def resolve_service_path(self, variant):
        """Resolve the path to optional services"""
        url = self.root + self.SERVICE_RESOLUTION_PATH + variant
        return self.opener.open(url).read()

    @property
    def release(self):
        """
        Returns the datawarehouse release
        =================================

        Service.release S{->} string

        The release is an arbitrary string used to distinguish
        releases of the datawarehouse. This usually coincides
        with updates to the data contained within. While a string,
        releases usually sort in ascending order of recentness
        (eg: "release-26", "release-27", "release-28"). They can also
        have less machine readable meanings (eg: "beta")

        @rtype: string
        """
        if self._release is None:
            with closing(self.opener.open(self.root + self.RELEASE_PATH)) as resp:
                self._release = ensure_str(resp.read()).strip()
        return self._release

    def load_query(self, xml, root=None):
        """
        Construct a new Query object for the given webservice
        =====================================================

        This is the standard method for instantiating new Query
        objects. Queries require access to the data model, as well
        as the service itself, so it is easiest to access them through
        this factory method.

        @return: L{intermine314.query.Query}
        """
        query_class, _ = _query_classes()
        return query_class.from_xml(xml, self.model, root=root)

    def select(self, *columns, **kwargs):
        """
        Construct a new Query object with the given columns selected.
        =============================================================

        As new_query, except that instead of a root class, a list of
        output column expressions are passed instead.
        """
        if "xml" in kwargs:
            return self.load_query(kwargs["xml"])
        query_class, _ = _query_classes()
        query = query_class(self.model, self)
        if len(columns) == 1:
            view = columns[0]
            if isinstance(view, Attribute):
                return query.select("%s.%s" % (view.declared_in.name, view))

            if isinstance(view, Reference):
                return query.select("%s.%s.*" % (view.declared_in.name, view))
            elif not isinstance(view, Column) and not str(view).endswith("*"):
                path = self.model.make_path(view)
                if not path.is_attribute():
                    return query.select(str(view) + ".*")

        return query.select(*columns)

    new_query = select

    def get_template(self, name):
        """
        Returns a template of the given name
        ====================================

        Tries to retrieve a template of the given name
        from the webservice. If you are trying to fetch
        a private template (ie. one you made yourself
        and is not available to others) then you may need to authenticate

        @see: L{intermine314.webservice.Service.__init__}

        @param name: the template's name
        @type name: string

        @raise ServiceError: if the template does not exist
        @raise QueryParseError: if the template cannot be parsed

        @return: L{intermine314.query.Template}
        """
        try:
            t = self.templates[name]
        except KeyError:
            raise ServiceError("There is no template called '" + name + "' at this service")
        _, template_class = _query_classes()
        if not isinstance(t, template_class):
            t = template_class.from_xml(t, self.model, self)
            self.templates[name] = t
        return t

    def get_template_by_user(self, name, username):
        """
        Returns a template of the given name belonging to username
        ==========================================================

        Tries to retrieve a template of the given name belonging
        to the username from the webservice. You need to authenticate
        as admin

        @see: L{intermine314.webservice.Service.__init__}

        @param name: the template's name
        @type name: string

        @param username: the username
        @type name: string

        @raise ServiceError: if the template or user does not exist
        @raise QueryParseError: if the template cannot be parsed

        @return: L{intermine314.query.Template}
        """
        try:
            templates = self.all_templates[username]
        except KeyError:
            raise ServiceError("There is no user called '" + username + "'")
        try:
            t = templates[name]
        except KeyError:
            raise ServiceError(
                "There is no template called '" + name + "' at this service belonging to '" + username + "'"
            )
        _, template_class = _query_classes()
        if not isinstance(t, template_class):
            t = template_class.from_xml(t, self.model, self)
            t.user_name = username
            self.all_templates[name] = t
        return t

    def _get_json(self, path, payload=None):
        headers = {"Accept": "application/json"}
        with closing(self.opener.open(self.root + path, payload, headers=headers)) as resp:
            data = json.loads(ensure_str(resp.read()))
            if data["error"] is not None:
                raise ServiceError(data["error"])
            return data

    def _get_xml(self, path):
        headers = {"Accept": "application/xml"}
        with closing(self.opener.open(self.root + path, headers=headers)) as sock:
            return minidom.parse(sock)

    def search(self, term, **facets):
        """
        Perform an unstructured search by term
        =======================================

        This seach method performs a search of all objects
        indexed by the service endpoint, returning results
        and facets for those results.

        @param term The search term
        @param facets The facets to search by (eg: Organism = 'H. sapiens')

        @return (list, dict) The results, and a dictionary of facetting informtation.
        """
        if isinstance(term, bytes):
            term = ensure_str(term)
        params = [("q", term)]
        for facet, value in list(facets.items()):
            if isinstance(value, bytes):
                value = ensure_str(value)
            params.append(("facet_{0}".format(facet), value))
        payload = urlencode(params, doseq=True)
        resp = self._get_json(self.SEARCH_PATH, payload=payload)
        return (resp["results"], resp["facets"])

    @property
    def widgets(self):
        """
        The dictionary of widgets from the webservice
        ==============================================

        The set of widgets available to a service does not
        change between releases, so they are cached.
        If you are running a long running process, you may
        wish to periodically dump the cache by calling
        L{Service.flush}, or simply get a new Service object.

        @return dict
        """
        if self._widgets is None:
            ws = self._get_json(self.WIDGETS_PATH)["widgets"]
            self._widgets = dict(([w["name"], w] for w in ws))
        return self._widgets

    def resolve_ids(self, data_type, identifiers, extra="", case_sensitive=False, wildcards=False):
        """
        Submit an Identifier Resolution Job
        ===================================

        Request that a set of identifiers be resolved to objects in
        the data store.

        @param data_type: The type of these identifiers (eg. 'Gene')
        @type data_type: String

        @param identifiers: The ids to resolve (eg. ['eve', 'zen', 'pparg'])
        @type identifiers: iterable of string

        @param extra: A disambiguating value (eg. "Drosophila melanogaster")
        @type extra: String

        @param case_sensitive: Whether to treat IDs case sensitively.
        @type case_sensitive: Boolean

        @param wildcards: Whether or not to interpret wildcards (eg: "eve*")
        @type wildcards: Boolean

        @return: {idresolution.Job} The job.
        """
        if self.version < 10:
            raise ServiceError("This feature requires API version 10+")
        if not data_type:
            raise ServiceError("No data-type supplied")
        if not identifiers:
            raise ServiceError("No identifiers supplied")

        data = json.dumps(
            {
                "type": data_type,
                "identifiers": list(identifiers),
                "extra": extra,
                "caseSensitive": case_sensitive,
                "wildCards": wildcards,
            }
        )
        text = self.opener.post_content(self.root + self.IDS_PATH, data, InterMineURLOpener.JSON)
        ret = json.loads(text)
        if ret["error"] is not None:
            raise ServiceError(ret["error"])
        if ret["uid"] is None:
            raise Exception("No uid found in " + ret)

        return idresolution.Job(self, ret["uid"])

    def flush(self):
        """
        Flushes any cached data.
        """
        self._list_manager.delete_temporary_lists()
        self._list_manager = ListManager(self)
        self._templates = None
        self._all_templates = None
        self._all_templates_names = None
        self._model = None
        self._version = None
        self._release = None
        self._widgets = None

    @property
    def templates(self):
        """
        The dictionary of templates from the webservice
        ===============================================

        Service.templates S{->} dict(intermine314.query.Template|string)

        For efficiency's sake, Templates are not parsed until
        they are required, and until then they are stored as XML
        strings. It is recommended that in most cases you would want
        to use L{Service.get_template}.

        You can use this property however to test for template existence though::

         if name in service.templates:
            template = service.get_template(name)

        @rtype: dict

        """
        if self._templates is None:
            templates = {}
            dom = self._get_xml(self.TEMPLATES_PATH)
            for e in dom.getElementsByTagName("template"):
                name = e.getAttribute("name")
                if name in templates:
                    raise ServiceError("Two templates with same name: " + name)
                else:
                    templates[name] = e.toxml()
            self._templates = templates
        return self._templates

    @property
    def all_templates(self):
        """
        The dictionary of templates by users from the webservice
        ========================================================

        Service.all_templates S{->} dict(string|string)

        You need to be authenticated as admin.

        For efficiency's sake, Templates are not parsed until
        they are required, and until then they are stored as XML
        strings. It is recommended that in most cases you would want
        to use L{Service.get_template}.

        You can use this property however to test for template existence::

         if name in service.templates:
           template = service.get_template(name)

        @rtype: dict

        """
        if self._all_templates is None:
            all_templates = {}
            dom = self._get_xml(self.ALL_TEMPLATES_PATH)
            for e in dom.getElementsByTagName("template"):
                user = e.getAttribute("userName")
                name = e.getAttribute("name")
                if user in all_templates:
                    templates = all_templates[user]
                    templates[name] = e.toxml()
                else:
                    templates = {}
                    templates[name] = e.toxml()
                    all_templates[user] = templates
            self._all_templates = all_templates
        return self._all_templates

    @property
    def all_templates_names(self):
        """
        The dictionary of templates names by users from the webservice
        =============================================================

        Service.all_templates_names S{->} dict(string|array)

        You need to be authenticated as admin.

        Example::
          allTemplatesNames = service.all_templates_names
          for user in allTemplatesNames:
            userTemplatesNames = allTemplatesNames[user]
            for templateName in userTemplatesNames:
              template = service.get_template_by_user(templateName, user)

        @rtype: dict

        """
        if self._all_templates_names is None:
            all_templates_names = {}
            dom = self._get_xml(self.ALL_TEMPLATES_PATH)
            for e in dom.getElementsByTagName("template"):
                user = e.getAttribute("userName")
                name = e.getAttribute("name")
                if user in all_templates_names:
                    all_templates_names[user].append(name)
                else:
                    templates_names = []
                    templates_names.append(name)
                    all_templates_names[user] = templates_names
            self._all_templates_names = all_templates_names
        return self._all_templates_names

    @property
    def model(self):
        """
        The data model for the webservice you are querying
        ==================================================

        Service.model S{->} L{intermine314.model.Model}

        This is used when constructing queries to provide them
        with information on the structure of the data model
        they are accessing. You are very unlikely to want to
        access this object directly.

        raises ModelParseError: if the model cannot be read

        @rtype: L{intermine314.model.Model}

        """
        if self._model is None:
            model_url = self.root + self.MODEL_PATH
            self._model = Model(model_url, self)
        return self._model

    def get_results(self, path, params, rowformat, view, cld=None):
        """
        Return an Iterator over the rows of the results
        ===============================================

        This method is called internally by the query objects
        when they are called to get results. You will not
        normally need to call it directly

        @param path: The resource path (eg: "/query/results")
        @type path: string
        @param params: The query parameters for this request as a dictionary
        @type params: dict
        @param rowformat: One of "rr", "object", "count", "dict", "list", "tsv", "csv", "jsonrows", "jsonobjects"
        @type rowformat: string
        @param view: The output columns
        @type view: list

        @raise WebserviceError: for failed requests

        @return: L{intermine314.webservice.ResultIterator}
        """
        return ResultIterator(self, path, params, rowformat, view, cld)

    @requires_version(9)
    def register(self, username, password):
        """
        Register a new user with this service.
        =======================================

        @return {Service} an authenticated service.
        """
        if isinstance(username, bytes):
            username = ensure_str(username)
        if isinstance(password, bytes):
            password = ensure_str(password)
        payload = urlencode({"name": username, "password": password})
        registrar = Service(self.root)
        resp = registrar._get_json(self.USERS_PATH, payload=payload)
        token = resp["user"]["temporaryToken"]
        return Service(self.root, token=token)

    @requires_version(16)
    def get_deregistration_token(self, validity=300):
        if validity < 1 or validity > 24 * 60 * 60:
            raise ValueError("Validity not a reasonable value: 1ms - 2hrs")
        params = urlencode({"validity": str(validity)})
        resp = self._get_json("/user/deregistration", payload=params)
        return resp["token"]

    @requires_version(16)
    def deregister(self, deregistration_token):
        """
        Remove a User from the service
        ==============================

        @param deregistration_token A token to prove you really want to do this

        @return string All the user's data.
        """
        if "uuid" in deregistration_token:
            deregistration_token = deregistration_token["uuid"]

        path = self.root + "/user"
        params = {"deregistrationToken": deregistration_token, "format": "xml"}
        uri = path + "?" + urlencode(params)
        self.flush()
        userdata = self.opener.delete(uri)
        return userdata



from io import StringIO
from urllib.request import urlopen


def openAnything(source):
    # Try to open with urllib (http, ftp, file url)
    try:
        return urlopen(source)
    except (ValueError, IOError, OSError):
        pass

    try:
        return open(source)
    except (IOError, OSError):
        pass

    return StringIO(str(source))


class ReadableException(Exception):
    def __init__(self, message, cause=None):
        self.message = message
        self.cause = cause

    def __str__(self):
        if self.cause is None:
            return repr(self.message)
        else:
            return repr(self.message) + repr(self.cause)



from __future__ import annotations

_POLARS_MODULE = None
_POLARS_CHECKED = False
_DUCKDB_MODULE = None
_DUCKDB_CHECKED = False


def optional_polars():
    global _POLARS_MODULE, _POLARS_CHECKED
    if not _POLARS_CHECKED:
        try:
            import polars as polars_module
        except ImportError:
            polars_module = None
        _POLARS_MODULE = polars_module
        _POLARS_CHECKED = True
    return _POLARS_MODULE


def require_polars(api_name):
    polars_module = optional_polars()
    if polars_module is None:
        raise ImportError(f"polars is required for {api_name}. Install with: pip install polars")
    return polars_module


def optional_duckdb():
    global _DUCKDB_MODULE, _DUCKDB_CHECKED
    if not _DUCKDB_CHECKED:
        try:
            import duckdb as duckdb_module
        except ImportError:
            duckdb_module = None
        _DUCKDB_MODULE = duckdb_module
        _DUCKDB_CHECKED = True
    return _DUCKDB_MODULE


def require_duckdb(api_name):
    duckdb_module = optional_duckdb()
    if duckdb_module is None:
        raise ImportError(f"duckdb is required for {api_name}. Install with: pip install duckdb")
    return duckdb_module


def quote_sql_string(value):
    return "'" + str(value).replace("'", "''") + "'"
